# [1] https://doi.org/10.1002/jcc.10156
#     Geometry optimization with QM/MM, ONIOM, and other combined methods.
#     I. Microiterations and constraints
#     Vreven, Morokuma, Farkas, Schlegel, Frisch, 2003
# [2] https://doi.org/10.1039/A909486E
#     Linear scaling geometry optimisation and transition state search
#     in hybrid delocalised internal coordinates
#     Billeter, Turner, Thiel, 2000
import logging
from pprint import pprint

import numpy as np

from pysisyphus.calculators import IPIServer, ONIOM
from pysisyphus.Geometry import Geometry
from pysisyphus.intcoords.exceptions import RebuiltInternalsException
from pysisyphus.helpers_pure import full_expand, highlight_text, recursive_update
from pysisyphus.optimizers.Optimizer import Optimizer


def get_geom_kwargs(layer_ind, layer_mask):
    coord_type = "tric" if layer_ind == 0 else "cartesian"
    all_indices = np.arange(layer_mask.size)
    freeze_atoms = all_indices[layer_mask]
    geom_kwargs = {
        "type": coord_type,
        "freeze_atoms": freeze_atoms,
        "coord_kwargs": {
            "freeze_atoms_exclude": layer_ind == 0,
        },
    }
    return geom_kwargs


def get_opt_kwargs(opt_key, layer_ind, thresh):
    # Some defaults tailored to LayerOpt
    opt_defaults = {
        "lbfgs": {
            "mu_reg": 0.1,
        },
        "plbfgs": {
            "precon_kind": "full_fast",
            "precon_update": 50,
        },
    }
    if layer_ind == 0:
        opt_kwargs = {
            "type": "rfo",
            "thresh": "never",
            # "prefix": "model",
            # "dump": True,
            # "h5_group_name": "model_opt",
        }
    else:
        if opt_key is None:
            opt_key = "lbfgs"

        opt_kwargs = {
            "type": opt_key,
            "max_cycles": 1_500,
            "thresh": thresh,
            "overachieve_factor": 3,
            # "prefix": f"layer_{layer_ind:02d}",
            # "dump": True,
        }
        try:
            opt_kwargs.update(opt_defaults[opt_key])
        except KeyError:
            pass
    return opt_kwargs


logger = logging.getLogger("optimizer")


class Layers:
    def __init__(self, geometry, opt_thresh, layers=None):
        self.geometry = geometry
        self.layers = layers
        self.opt_thresh = opt_thresh

        print(highlight_text("Layers", level=1))
        pprint(self.layers, compact=True)
        print("")

        atoms = geometry.atoms
        all_indices = np.arange(len(geometry.atoms))
        # Boolean array; every item corresponds to an atom in the total system.
        freeze_mask = np.full_like(all_indices, True, dtype=bool)

        self.indices = list()
        self.geom_getters = list()
        self.opt_getters = list()

        # We import 'get_opt_cls' in the constructor, as it relies on dicts, that
        # contain references to this class (LayerOpt). Trying to import 'get_opt_cls'
        # at the top of the module will result in an circular import.
        from pysisyphus.optimizers.cls_map import get_opt_cls

        indices_below = set()
        # Iterate in reverse order from smallest (lowest) layer to biggest (highest) layer.
        # to setup geom_getter and opt_getter.
        for i, layer in enumerate(layers[::-1]):
            try:
                indices = full_expand(layer["indices"])
            # Allow missing 'indices' key. Then it is assumed that this layer contains the
            # whole system.
            except KeyError:
                assert (
                    i != 0
                ), "Found whole system in highest level layer. I don't like that!"
                indices = all_indices
            # Drop indices from layers below ...
            indices = sorted(set(indices) - indices_below)
            # ... and update 'indices_below' for the next layer
            indices_below.update(set(indices))
            self.indices.append(indices)
            # Set up mask that indicates which atoms to freeze in the current layer (
            # all atoms that are not in the current layer.)
            layer_mask = freeze_mask.copy()
            layer_mask[indices] = False

            try:
                try:
                    calc_kwargs = {
                        "address": layer["address"],
                    }
                except KeyError:
                    _calc_kwargs = layer["calc"]
                    calc_kwargs = recursive_update({}, _calc_kwargs)
                    # The popped calc key is currently unused as using an IPIServer is
                    # mandatory.
                    _ = calc_kwargs.pop("type", None)
                calc = IPIServer(**calc_kwargs)
                # When calculating layer 0 we have access to the true energy of the system.
                # So we assign the calculator of layer0 to the actual geometry containing
                # the whole system.
                if i == 0:
                    geometry.set_calculator(calc)
            # If no address is given, we assume that pysisyphus' ONIOM calculator
            # is used.
            except KeyError as err:
                try:
                    calc = layer["layer_calc"]
                except KeyError:
                    print(
                        "Currently, a socket address for an IPI-protol client is mandatory!"
                    )
                    raise err

            ####################
            #  Geometry setup  #
            ####################

            _geom_kwargs = layer.get("geom", dict())
            _geom_kwargs = recursive_update({}, _geom_kwargs)
            geom_kwargs = get_geom_kwargs(i, layer_mask=layer_mask)
            try:
                geom_kwargs.update(_geom_kwargs)
            # Allow empty "geom:" block
            except TypeError:
                pass
            coord_type = geom_kwargs.pop("type")

            # Geometry
            def get_geom_getter(persistent_geom=None):
                # Rebind the variables here, otherwise the wrong geom_kwargs
                # and calc will be used, as they are redefined in the next loop cycle.
                layer_geom_kwargs = geom_kwargs.copy()
                layer_calc = calc

                def get_geom(coords3d):
                    if persistent_geom is not None:
                        geom = persistent_geom
                        geom.coords3d = coords3d
                    else:
                        geom = Geometry(
                            atoms,
                            coords3d.copy(),
                            coord_type=coord_type,
                            **layer_geom_kwargs,
                        )
                        geom.set_calculator(layer_calc)
                    return geom

                return get_geom

            get_geom = get_geom_getter()
            # Use a persistent Geometry for the layer 0. Overwrite the function
            # above with a definition that always returns the same geometry.
            if i == 0:
                geom0 = get_geom(geometry.coords3d)
                get_geom = get_geom_getter(geom0)

            self.geom_getters.append(get_geom)

            #####################
            #  Optimizer setup  #
            #####################

            _opt_kwargs = layer.get("opt", dict())
            _opt_kwargs = recursive_update({}, _opt_kwargs)
            opt_key = _opt_kwargs.get("type", None)
            opt_kwargs = get_opt_kwargs(opt_key, i, thresh=self.opt_thresh)
            try:
                opt_kwargs.update(_opt_kwargs)
            # Allow empty "opt:" block
            except TypeError:
                pass
            opt_key = opt_kwargs.pop("type")
            opt_cls = get_opt_cls(opt_key)

            def get_opt_getter():
                layer_opt_kwargs = opt_kwargs
                layer_opt_cls = opt_cls

                def get_opt(geom, **kwargs):
                    kwargs_ = layer_opt_kwargs.copy()
                    kwargs_.update(kwargs)
                    opt = layer_opt_cls(geom, **kwargs_)
                    return opt

                return get_opt

            get_opt = get_opt_getter()
            # Persistent optimizer for most expensive layer.
            if i == 0:
                model_opt = get_opt(geom0)

                def get_opt(geom, **kwargs):
                    return model_opt

            self.opt_getters.append(get_opt)

        # Most expensive layer will come last.
        self.indices = self.indices[::-1]
        self.geom_getters = self.geom_getters[::-1]
        self.opt_getters = self.opt_getters[::-1]

    @classmethod
    def from_oniom_calculator(cls, geometry, oniom_calc=None, layers=None, **kwargs):
        calc = geometry.calculator
        if calc is None:
            calc = oniom_calc

        if layers is not None:
            assert len(layers) == len(calc.layers), (
                f"ONIOM calculator has {len(calc.layers)} layers, but only "
                f"{len(layers)} layer were defined in 'layers:'!"
            )

            layers = list(layers)
            for i, layer in enumerate(layers):
                if layer is None:
                    layers[i] = dict()
                elif isinstance(layer, dict):
                    pass
                else:
                    raise Exception("Layer definition must be empty or dict-like!")
        else:
            layers = [dict() for _ in calc.layers]

        for i, layer in enumerate(calc.layers):
            assert len(layer) == 1, "Multicenter-ONIOM is not yet supported!"
            model = layer[0]
            link_hosts = [link.parent_ind for link in model.links]
            indices = model.atom_inds + link_hosts
            layer_calc = calc.get_layer_calc(i)
            layer = {
                "layer_calc": layer_calc,
                "indices": indices,
            }
            layers[i].update(layer)
        return Layers(geometry, layers=layers, **kwargs)

    def __len__(self):
        return len(self.layers)


class LayerOpt(Optimizer):
    def __init__(
        self,
        geometry: Geometry,
        layers: dict = None,
        **kwargs,
    ) -> None:
        super().__init__(geometry, **kwargs)
        assert geometry.coord_type in ("cart", "cartesian")

        layers_kwargs = {
            "geometry": self.geometry,
            "opt_thresh": self.thresh,
            "layers": layers,
        }
        # Construct layers from ONIOM calculator
        if isinstance(self.geometry.calculator, ONIOM):
            layers = Layers.from_oniom_calculator(**layers_kwargs)
        else:
            layers = Layers(**layers_kwargs)
        self.layers = layers

        self.micro_cycles = list()
        self.micro_cycles_converged = list()

    @property
    def layer_num(self) -> int:
        return len(self.layers)

    @property
    def model_opt(self):
        """Return the persistent optimizer belonging to the model system.
        We don't have to supply any coordinates to the optimizer of the
        most expensive layer, as it is persistent, as well as the associated geometry."""
        return self.layers.opt_getters[-1](None)

    def optimize(self) -> None:
        coords3d_org = self.geometry.coords3d.copy()
        coords3d_cur = coords3d_org.copy()
        cur_micro_cycles = list()
        cur_micro_cycles_converged = list()
        for i, (indices, get_geom, get_opt) in enumerate(
            zip(self.layers.indices, self.layers.geom_getters, self.layers.opt_getters)
        ):
            print(highlight_text(f"Layer {i}", level=1))
            is_last_layer = i == self.layer_num - 1
            geom = get_geom(coords3d_cur)
            prefix = f"mc{self.cur_cycle:03d}"
            opt = get_opt(geom, prefix=prefix, h5_group_name=f"{prefix}_opt")
            if is_last_layer:
                if self.cur_cycle == 0:
                    opt.prepare_opt()
                break
            opt.run()
            cur_micro_cycles.append(opt.cur_cycle + 1)
            cur_micro_cycles_converged.append(opt.is_converged)
            coords3d_cur[indices] = geom.coords3d[indices]
        self.micro_cycles.append(cur_micro_cycles)
        self.micro_cycles_converged.append(cur_micro_cycles_converged)

        ####################
        # Relax last layer #
        ####################

        # 'geom' and 'indices' for the last layer were defined in the for-loop
        # above, before breaking from the loop.
        #
        # Calculate forces and energy of the last layer. This have to be the "true"
        # ONIOM forces of the system, containing all contributions. That's why we
        # also save them as the true forces in the optimizer.
        cart_forces = geom.cart_forces
        energy = geom.energy
        self.energies.append(energy)
        self.forces.append(cart_forces.copy())

        # Also store relevant quantities in the optimizer of the last layer, so
        # things like Hessian updates are possible. These quantities are usually
        # stored in the big optimization-loop in Optimizer.run(). As run() is
        # never called for the last optimizer we have to store them manually.
        opt.coords.append(geom.coords.copy())
        opt.cart_coords.append(geom.cart_coords.copy())

        # Calculate one step
        opt.cur_cycle = self.cur_cycle
        int_step = opt.optimize()
        opt.steps.append(int_step)
        try:
            geom.coords = geom.coords + int_step
        except RebuiltInternalsException:
            geom.reset_coords()
            opt.reset()
        coords3d_cur[indices] = geom.coords3d[indices]

        full_step = coords3d_cur - coords3d_org
        return full_step.flatten()

    def check_convergence(self, *args, **kwargs):
        """Check if we must use the model optimizer to signal convergence."""
        opt = self.model_opt
        if opt.thresh != "never":
            return opt.check_convergence()
        else:
            return super().check_convergence(*args, **kwargs)

    def print_opt_progress(self, *args, **kwargs):
        """Pick the correct method to report opt_progress.

        When the model optimizer decides convergence we also report optimization
        progress using its data and not the data from LayerOpt, where the total
        ONIOM gradient is stored."""
        opt = self.model_opt
        if opt.thresh != "never":
            func = opt.print_opt_progress
        else:
            func = super().print_opt_progress
        func(*args, **kwargs)

    def postprocess_opt(self) -> None:
        coord_types = list()
        for layer in self.layers.layers:
            try:
                coord_type = layer["geom"]["type"]
                coord_types.append(coord_type)
            except KeyError:
                pass
        micro_sum = np.array(self.micro_cycles).sum()
        print("\nMicrocycles:")
        print("\t", end="")
        pprint(self.micro_cycles)
        print(f"\t@@@ Σ {micro_sum},", ", ".join(coord_types))
        print(f"\t@@@ Macrocycles: {self.cur_cycle+1}, converged? {self.is_converged}")
        print("\t@@@")



# [1] http://link.aip.org/link/doi/10.1063/1.1515483
#     The efficient optimization of molecular geometries using redundant internal
#     coordinates
#     Bakken, Helgaker, 2002

from collections import namedtuple
import itertools as it
from typing import Optional

import numpy as np
from scipy.spatial.distance import pdist, squareform
from sklearn.cluster import KMeans

from pysisyphus.constants import BOHR2ANG
from pysisyphus.config import BEND_MIN_DEG, DIHED_MAX_DEG
from pysisyphus.helpers_pure import log, sort_by_central, merge_sets
from pysisyphus.elem_data import VDW_RADII, COVALENT_RADII as CR
from pysisyphus.intcoords import Stretch, Bend, LinearBend, Torsion
from pysisyphus.intcoords.setup_fast import find_bonds as find_bonds_fast
from pysisyphus.intcoords.PrimTypes import PrimTypes, PrimMap, Rotations
from pysisyphus.intcoords.valid import bend_valid, dihedral_valid


BOND_FACTOR = 1.3


def get_pair_covalent_radii(atoms):
    atoms = [a.lower() for a in atoms]
    cov_radii = np.array([CR[a.lower()] for a in atoms])
    pair_cov_radii = np.array([r1 + r2 for r1, r2 in it.combinations(cov_radii, 2)])
    return pair_cov_radii


def get_bond_mat(geom, bond_factor=BOND_FACTOR):
    cdm = pdist(geom.coords3d)
    pair_cov_radii = get_pair_covalent_radii(geom.atoms)
    bond_mat = squareform(cdm <= (pair_cov_radii * bond_factor))
    return bond_mat


def get_bond_sets(
    atoms, coords3d, bond_factor=BOND_FACTOR, return_cdm=False, return_cbm=False
):
    """I'm sorry, but this function does not return sets, but an int ndarray."""
    cdm = pdist(coords3d)
    # Generate indices corresponding to the atom pairs in the
    # condensed distance matrix cdm.
    atom_inds = list(it.combinations(range(len(coords3d)), 2))
    atom_inds = np.array(atom_inds, dtype=int)
    scaled_cr_sums = bond_factor * get_pair_covalent_radii(atoms)
    # condensed bond matrix
    cbm = cdm <= scaled_cr_sums
    bond_inds = atom_inds[cbm]
    if not return_cbm and not return_cdm:
        return bond_inds
    add_returns = tuple(
        [mat for flag, mat in ((return_cdm, cdm), (return_cbm, cbm)) if flag]
    )
    return (bond_inds,) + add_returns


def get_fragments(atoms, coords, bond_inds=None, bond_factor=BOND_FACTOR):
    """This misses unconnected single atoms!"""
    coords3d = coords.reshape(-1, 3)
    if bond_inds is None:
        # Bond indices without interfragment bonds and/or hydrogen bonds
        bond_inds = get_bond_sets(atoms, coords3d, bond_factor=bond_factor)

    bond_ind_sets = [frozenset(bi) for bi in bond_inds]
    fragments = merge_sets(bond_ind_sets)

    return fragments


def connect_fragments(
    cdm, fragments, max_aux=3.78, aux_factor=BOND_FACTOR, logger=None
):
    """Determine the smallest interfragment bond for a list
    of fragments and a condensed distance matrix. For more than a few fragments
    this function performs poorly, as each fragment is connected to all reamaining
    fragments, leading to an explosion of bonds, bends and dihedrals."""
    if len(fragments) > 1:
        log(
            logger,
            f"Detected {len(fragments)} fragments. Generating interfragment bonds.",
        )
    dist_mat = squareform(cdm)
    interfrag_inds = list()
    aux_interfrag_inds = list()
    for frag1, frag2 in it.combinations(fragments, 2):
        log(logger, f"\tConnecting {len(frag1)} atom and {len(frag2)} atom fragment")
        inds = [(i1, i2) for i1, i2 in it.product(frag1, frag2)]
        distances = np.array([dist_mat[ind] for ind in inds])

        # Determine minimum distance bond
        min_ind = distances.argmin()
        min_dist = distances[min_ind]
        interfrag_bond = tuple(inds[min_ind])
        interfrag_inds.append(interfrag_bond)
        log(logger, f"\tMinimum distance bond: {interfrag_bond}, {min_dist:.4f} au")

        # Determine auxiliary interfragment bonds that are either below max_aux
        # (default 2 Å, ≈ 3.78 au), or less than aux_factor (default 1.3) times the
        # minimum interfragment distance.
        below_max_aux = [
            ind for ind in inds if (dist_mat[ind] < max_aux) and (ind != interfrag_bond)
        ]
        if below_max_aux:
            log(
                logger,
                f"\tAux. interfrag bonds below {max_aux*BOHR2ANG:.2f} Å:\n"
                + "\n".join(
                    [f"\t\t{ind}: {dist_mat[ind]:.4f} au" for ind in below_max_aux]
                ),
            )
        scaled_min_dist = aux_factor * min_dist
        above_min_dist = [
            ind
            for ind in inds
            if (dist_mat[ind] < scaled_min_dist)
            and (ind != interfrag_bond)
            and (ind not in below_max_aux)
        ]
        if above_min_dist:
            log(
                logger,
                f"\tAux. interfrag bonds below {aux_factor:.2f} * min_dist:\n"
                + "\n".join(
                    [f"\t\t{ind}: {dist_mat[ind]:.4f} au" for ind in above_min_dist]
                ),
            )
        aux_interfrag_inds.extend(below_max_aux)
        aux_interfrag_inds.extend(above_min_dist)
    # Or as Philipp proposed: two loops over the fragments and only
    # generate interfragment distances. So we get a full matrix with
    # the original indices but only the required distances.
    return interfrag_inds, aux_interfrag_inds


def connect_fragments_kmeans(
    cdm,
    fragments,
    atoms,
    aux_below_thresh=3.7807,  # 2 Å
    aux_add_dist=2.8356,  # 1.5 Å
    aux_keep=5,
    aux_no_hh=True,
    min_dist_thresh=5.0,
    min_scale=1.2,
    logger=None,
):
    """Generate (auxiliary) interfragment bonds.

    In the a first step, minimum distance interfragment bonds (IFBs) are determined between
    all possible fragment pairs. Similarly, possible auxiliary IFBs are determined.
    Candidates for auxiliary IFBs are:
        IFB <= aux_below_thresh, default 2 Å
        IFB <= (minimum distance IFB + aux_add_dist), default 1.5 Å
    By default, only the first aux_keep (default = 5) auxiliary IFBs are kept.

    Connecting all fragments can lead to bonds between very distant atoms. If more than
    two fragments are present we cluster the minimum distance IFB distances using KMeans,
    to determine a reasonable length for valid IFBs. We start out with two clusters and
    increase the number of cluster until the center of one cluster is around the scaled
    global minimum distance between the fragments. The center of this cluster is then used
    as a cutoff vor valid IFBs.

    After pruning all possible IFBs we can determine the fragment pairs, that are actually
    connected. This information is then used to also prune possible interfragment bonds.
    Only auxiliary IFBs between fragments that are actually connected via IFBs are kept.
    """
    # atoms = [atom.lower() for atom in atoms]
    if len(fragments) > 1:
        log(
            logger,
            f"Detected {len(fragments)} fragments. Generating interfragment bonds.",
        )
    dist_mat = squareform(cdm)

    frag_pairs = list()
    interfrag_inds = list()
    interfrag_dists = list()
    aux_dict = dict()
    for i, j in it.combinations(range(len(fragments)), 2):
        frag1 = fragments[i]
        frag2 = fragments[j]
        log(logger, f"\tConnecting {len(frag1)}-atom and {len(frag2)}-atom fragments.")
        # Pairs of possible interfragment bonds
        inds = np.array([(i1, i2) for i1, i2 in it.product(frag1, frag2)], dtype=int)
        distances = np.array([dist_mat[k, l] for k, l in inds])

        frag_pairs.append((i, j))
        # Determine minimum distance bond
        min_ind = distances.argmin()
        min_dist = distances[min_ind]
        interfrag_inds.append(inds[min_ind])
        interfrag_dists.append(min_dist)
        """
        But also consider other bonds with reasonable lengths as auxiliary interfrag bonds.
        Contrary to [1] we don't use a scaled minimum interfragment distances (MID),
        but just add a fixed value to the MID. Scaling a possibly big MID would
        probably include too many coordinates.
        """
        # if aux_no_hh:
        # is_hh = np.array(
        # [(atoms[k] == "h") and (atoms[k] == atoms[l]) for k, l in inds]
        # )
        # else:
        # is_hh = np.zeros(len(inds))

        aux_mask = np.logical_or(
            distances <= aux_below_thresh,
            distances <= (min_dist + aux_add_dist),
        )
        # aux_mask = distances <= 1.1 * min_dist
        # aux_mask = np.logical_and(aux_mask, ~is_hh)
        # Don't include interfragment bond
        aux_mask[min_ind] = False
        aux_inds = inds[aux_mask]

        if aux_keep >= 0:
            aux_dists = distances[aux_mask]
            aux_keep_inds = aux_dists.argsort()[:aux_keep]
            aux_inds = aux_inds[aux_keep_inds]

        aux_dict[(i, j)] = aux_inds
    frag_pairs = np.array(frag_pairs, dtype=int)

    """
    The code below tries to determine a reasonable distance, to define
    interfragment bonds. We don't want to use all interfragment bonds defined
    above, as this would connect all fragments to each other, resulting in an
    explosion of the number of internal coordinates.

    Here we try to cluster the interfragment distances, until one cluster center
    comes close to the scaled minimum interfragment distance. We then use this
    distance to filter all interfragment bonds.
    """
    if len(fragments) > 2:
        dists = np.reshape(interfrag_dists, (-1, 1))
        min_dist = dists.min()

        for n_clusters in range(2, 10):
            kmeans = KMeans(n_clusters=n_clusters)
            _ = kmeans.fit_predict(dists)
            min_center = kmeans.cluster_centers_.min()

            if min_center <= (min_scale * min_dist):
                break
        else:
            raise Exception("Not handled!")

        interfrag_inds = np.array(interfrag_inds)
        mask = dists <= min_scale * min_center
        # We also use interfragment bonds that still have a reasonable length.
        if (min_dist_thresh is not None) and (min_dist_thresh > 0.0):
            mask |= dists <= min_dist_thresh
        mask = mask.flatten()
        interfrag_inds = interfrag_inds[mask]
        conn_frags_mask = mask
    else:
        conn_frags_mask = np.ones(len(frag_pairs), dtype=bool)

    # Only keep auxiliary interfragment bonds between actually connected fragments
    actually_connected_frags = frag_pairs[conn_frags_mask]
    aux_interfrag_inds = list(
        it.chain(*[aux_dict[(i, j)] for i, j in actually_connected_frags])
    )
    aux_interfrag_inds = np.array(aux_interfrag_inds, dtype=int)
    return interfrag_inds, aux_interfrag_inds


def connect_fragments_ahlrichs(
    cdm,
    fragments,
    atoms,
    min_dist_scale=1.1,
    scale=1.2,
    avoid_h=False,
    logger=None,
):
    atoms = [atom.lower() for atom in atoms]
    if len(fragments) > 1:
        log(
            logger,
            f"Detected {len(fragments)} fragments. Generating interfragment bonds.",
        )
    dist_mat = squareform(cdm)

    frag_pairs = list()
    interfrag_inds = list()
    max_dist = 3.0 / BOHR2ANG
    all_fragment_inds = [i for i, _ in enumerate(fragments)]
    # Leave out the first fragment, so we can also handle only 1 fragment here.
    unconnected_fragment_inds = all_fragment_inds.copy()[1:]
    h_inds = set([i for i, atom in enumerate(atoms) if atom.lower() == "h"])
    while True:
        for i, j in it.product(all_fragment_inds, unconnected_fragment_inds):
            # Don't try to connect a fragment to itself and don't try to connect
            # fragments two times, e.g., (0 and 2) as well as (2 and 0).
            if i >= j:
                continue

            frag1 = fragments[i]
            frag2 = fragments[j]
            log(
                logger,
                f"\tConnecting {len(frag1)}-atom and {len(frag2)}-atom fragments.",
            )
            # Pairs of possible interfragment bonds
            inds = np.array(
                [(i1, i2) for i1, i2 in it.product(frag1, frag2)], dtype=int
            )
            distances = np.array([dist_mat[k, l] for k, l in inds])

            frag_pairs.append((i, j))
            # Determine minimum distance bond
            min_ind = distances.argmin()
            min_dist = distances[min_ind]

            if avoid_h:
                sort_inds = np.argsort(distances, kind="stable")
                # Try to avoid interfragment bonds involving hydrogen.
                for (k, dist) in zip(sort_inds, distances[sort_inds]):
                    if set(inds[k]) & h_inds:
                        continue
                    if dist >= 1.5 * min_dist:
                        break
                    min_ind = k
                    min_dist = distances[k]
                    break

            if min_dist <= max_dist:
                # Scaling of bigger 'min_dist' values by a fixed factor can lead
                # to definition of many additional bonds. We restrict the offset
                # to at most 1 Å.
                offset = min((min_dist_scale - 1.0) * min_dist, 1.0 / BOHR2ANG)
                mask = distances <= (min_dist + offset)
                interfrag_inds.extend(inds[mask])
                # Indicate that the just connected fragments don't have to be
                # connected anymore. In the current cycle of the while loop additional
                # bonds to the just connected fragments can still be defined.
                unconnected_fragment_inds = [
                    k for k in unconnected_fragment_inds if k not in (i, j)
                ]
        # Leave the outer while loop when all fragments are connected
        if len(unconnected_fragment_inds) == 0:
            break
        # If there are still unconnected fragments present we allow longer
        # interfragment bonds.
        max_dist *= scale

    interfrag_inds = np.array(interfrag_inds, dtype=int)
    return interfrag_inds, list()


def get_hydrogen_bond_inds(atoms, coords3d, bond_inds, logger=None):
    tmp_sets = [frozenset(bi) for bi in bond_inds]
    hydrogen_inds = [i for i, a in enumerate(atoms) if a.lower() == "h"]
    x_inds = [i for i, a in enumerate(atoms) if a.lower() in "n o f p s cl".split()]
    hydrogen_bond_inds = list()
    for h_ind, x_ind in it.product(hydrogen_inds, x_inds):
        as_set = set((h_ind, x_ind))
        if as_set not in tmp_sets:
            continue
        # Check if distance of H to another electronegative atom Y is
        # greater than (1.1 * sum of their covalent radii) but smaller than
        # (0.9 * sum of their van der Waals radii). If the
        # angle X-H-Y is greater than 90° a hydrogen bond is asigned.
        y_inds = set(x_inds) - set((x_ind,))
        for y_ind in y_inds:
            y_atom = atoms[y_ind].lower()
            cov_rad_sum = CR["h"] + CR[y_atom]
            distance = Stretch._calculate(coords3d, (h_ind, y_ind))
            vdw_rad_sum = VDW_RADII["h"] + VDW_RADII[y_atom]
            angle = Bend._calculate(coords3d, (x_ind, h_ind, y_ind))
            if (1.1 * cov_rad_sum < distance < 0.9 * vdw_rad_sum) and (
                angle > np.pi / 2
            ):
                hydrogen_bond_inds.append((h_ind, y_ind))
                log(
                    logger,
                    f"Detected hydrogen bond between atoms {h_ind} "
                    f"({atoms[h_ind]}) and {y_ind} ({atoms[y_ind]})",
                )

    return hydrogen_bond_inds


def get_hydrogen_bond_inds_v2(atoms, coords3d, bond_inds, logger=None):
    def to_set(iterable):
        return {frozenset(_) for _ in iterable}

    atoms_lower = [atom.lower() for atom in atoms]
    # Determine Hydrogen indices
    org_h_inds = {i for i, atom in enumerate(atoms_lower) if atom == "h"}

    org_bond_sets = to_set(bond_inds)
    # Determine Hydrogen bonding partners in original bond set
    org_h_partners = dict()
    for org_bond in org_bond_sets:
        if h_set := org_bond & org_h_inds:
            (x_ind,) = org_bond - h_set
            (h_ind,) = h_set
            org_h_partners.setdefault(h_ind, list()).append(x_ind)

    hx = {"h", "n", "o", "f", "p", "s", "cl"}
    # Determine indices of potential hydrogen bond acceptors and hydrogen to
    # carry out a search for bonds with a bigger bond factor.
    hx_inds = [i for i, atom in enumerate(atoms) if atom.lower() in hx]
    hx_atoms = [atoms[i] for i in hx_inds]
    hx_map = {j: i for j, i in enumerate(hx_inds)}
    # See pysisyphus.elemdata.HBOND_FACTORS
    hx_coords3d = coords3d[hx_inds]
    # Search bonds with KDTree
    h_bonds = find_bonds_fast(hx_atoms, hx_coords3d, bond_factor=2.3)
    h_bonds = to_set(h_bonds)
    # Map back to original indices. Until now the indices were only in the basis
    # of the reduced number of atoms.
    h_bonds = {frozenset((hx_map[i], hx_map[j])) for i, j in h_bonds}
    # Drop h_bonds that were already defined as "normal" bonds
    h_bonds = h_bonds - org_bond_sets

    hydrogen_bond_inds = list()
    # h_bonds can still contain XX and HH bonds
    for h_bond in h_bonds:
        hi = h_bond & org_h_inds
        # Skip XX and HH. We are only interest in 'h_bond' with one hydrogen.
        if len(hi) in (0, 2):
            continue
        (h_ind,) = hi
        (y_partner,) = h_bond - hi
        v = coords3d[y_partner] - coords3d[h_ind]
        for x_partner in org_h_partners[h_ind]:
            u = coords3d[x_partner] - coords3d[h_ind]
            # > 90°
            if u.dot(v) < 0.0:
                hydrogen_bond_inds.append((h_ind, y_partner))
    return hydrogen_bond_inds


def get_bend_inds(coords3d, bond_inds, min_deg, max_deg, logger=None):
    bond_sets = {frozenset(bi) for bi in bond_inds}

    bend_inds = list()
    for bond_set1, bond_set2 in it.combinations(bond_sets, 2):
        union = bond_set1 | bond_set2
        if len(union) == 3:
            indices, _ = sort_by_central(bond_set1, bond_set2)
            if not bend_valid(coords3d, indices, min_deg, max_deg):
                log(logger, f"Bend {indices} is not valid!")
                continue
            bend_inds.append(indices)

    return bend_inds


def get_linear_bend_inds(coords3d, cbm, bends, min_deg=175, max_bonds=4, logger=None):
    linear_bends = list()
    complements = list()

    if min_deg is None:
        return linear_bends, complements

    bm = squareform(cbm)
    for bend in bends:
        deg = np.rad2deg(Bend._calculate(coords3d, bend))
        bonds = sum(bm[bend[1]])
        if (deg >= min_deg) and (bonds <= max_bonds):
            log(
                logger,
                f"Bend {bend}={deg:.1f}° is (close to) linear. "
                "Creating linear bend & complement.",
            )
            linear_bends.append(bend)
            complements.append(bend)
    return linear_bends, complements


def get_dihedral_inds(coords3d, bond_inds, bend_inds, max_deg, logger=None):
    max_rad = np.deg2rad(max_deg)
    bond_dict = dict()
    for from_, to_ in bond_inds:
        bond_dict.setdefault(from_, list()).append(to_)
        bond_dict.setdefault(to_, list()).append(from_)
    proper_dihedral_inds = list()
    improper_candidates = list()
    improper_dihedral_inds = list()

    def log_dihed_skip(inds):
        log(
            logger,
            f"Skipping generation of dihedral {inds} "
            "as some of the the atoms are (close too) linear.",
        )

    def set_dihedral_index(dihedral_ind, proper=True):
        dihed = tuple(dihedral_ind)
        check_in = proper_dihedral_inds if proper else improper_dihedral_inds
        # Check if this dihedral is already present
        if (dihed in check_in) or (dihed[::-1] in check_in):
            return
        # Assure that the angles are below 175° (3.054326 rad)
        if not dihedral_valid(coords3d, dihedral_ind, deg_thresh=max_deg):
            log_dihed_skip(dihedral_ind)
            return
        if proper:
            proper_dihedral_inds.append(dihed)
        else:
            improper_dihedral_inds.append(dihed)

    for bond, bend in it.product(bond_inds, bend_inds):
        central = bend[1]
        bend_set = set(bend)
        bond_set = set(bond)
        # Check if the two sets share one common atom. If not continue.
        intersect = bend_set & bond_set
        if len(intersect) != 1:
            continue

        # TODO: check collinearity of bond and bend.

        # When the common atom between bond and bend is a terminal, and not a central atom
        # in the bend we create a proper dihedral. Improper dihedrals are only created
        # when no proper dihedrals have been found.
        if central not in bond_set:
            # The new terminal atom in the dihedral is the one, that doesn' intersect.
            terminal = tuple(bond_set - intersect)[0]
            intersecting_atom = tuple(intersect)[0]
            bend_terminal = tuple(bend_set - {central} - intersect)[0]

            bend_rad = Bend._calculate(coords3d, bend)
            # Bend atoms are nearly collinear. Check if we can skip the central bend atom
            # and use an atom that is conneced to the terminal atom of the bend or bond.
            if bend_rad >= max_rad:
                bend_terminal_bonds = set(bond_dict[bend_terminal]) - bend_set
                bond_terminal_bonds = set(bond_dict[terminal]) - bond_set
                set_dihedrals = [
                    (terminal, intersecting_atom, bend_terminal, betb)
                    for betb in bend_terminal_bonds
                ] + [
                    (bend_terminal, intersecting_atom, terminal, botb)
                    for botb in bond_terminal_bonds
                ]
                # Hardcoded for now ... look ahead to next shell of atoms
                if not any(
                    [
                        dihedral_valid(coords3d, inds, deg_thresh=max_deg)
                        for inds in set_dihedrals
                    ]
                ):
                    set_dihedrals = []
                    for betb in bend_terminal_bonds:
                        bend_terminal_bonds_v2 = (
                            set(bond_dict[betb]) - bend_set - bond_set
                        )
                        set_dihedrals = [
                            (terminal, intersecting_atom, betb, betb_v2)
                            for betb_v2 in bend_terminal_bonds_v2
                        ]
                    for botb in bond_terminal_bonds:
                        bond_terminal_bonds_v2 = (
                            set(bond_dict[botb]) - bend_set - bond_set
                        )
                        set_dihedrals = [
                            (bend_terminal, intersecting_atom, botb, botb_v2)
                            for botb_v2 in bond_terminal_bonds_v2
                        ]
            elif intersecting_atom == bend[0]:
                set_dihedrals = [[terminal] + list(bend)]
            else:
                set_dihedrals = [list(bend) + [terminal]]
            [set_dihedral_index(dihed) for dihed in set_dihedrals]
        # If the common atom is the central atom we try to form an out
        # of plane bend / improper torsion. They may be created later on.
        else:
            fourth_atom = list(bond_set - intersect)
            dihedral_ind = list(bend) + fourth_atom
            # This way dihedrals may be generated that contain linear
            # atoms and these would be undefinied. So we check for this.
            if dihedral_valid(coords3d, dihedral_ind, deg_thresh=max_deg):
                improper_candidates.append(dihedral_ind)
            else:
                log_dihed_skip(dihedral_ind)

    # Now try to create the remaining improper dihedrals.
    if (len(coords3d) >= 4) and (len(proper_dihedral_inds) == 0):
        log(
            logger,
            "Could not define any proper dihedrals! Generating improper dihedrals!",
        )
        for improp in improper_candidates:
            set_dihedral_index(improp, proper=False)
        log(
            logger,
            "Permutational symmetry not considerd in generation of "
            "improper dihedrals.",
        )

    return proper_dihedral_inds, improper_dihedral_inds


def sort_by_prim_type(to_sort=None):
    if to_sort is None:
        to_sort = list()

    by_prim_type = [[], [], []]
    for item in to_sort:
        len_ = len(item)
        # len -> index
        #   2 ->     0 (bond)
        #   3 ->     1 (bend)
        #   4 ->     2 (torsion)
        by_prim_type[len_ - 2].append(tuple(item))
    return by_prim_type


CoordInfo = namedtuple(
    "CoordInfo",
    "bonds hydrogen_bonds interfrag_bonds aux_interfrag_bonds "
    "bends linear_bends linear_bend_complements "
    # "dihedrals typed_prims fragments cdm cbm".split(),
    "proper_dihedrals improper_dihedrals "
    "translation_inds rotation_inds cartesian_inds "
    "typed_prims fragments".split(),
)


def setup_redundant(
    atoms,
    coords3d,
    factor=BOND_FACTOR,
    define_prims=None,
    min_deg=BEND_MIN_DEG,
    dihed_max_deg=DIHED_MAX_DEG,
    lb_min_deg=None,
    lb_max_bonds=4,
    min_weight=None,
    tric=False,
    hybrid=False,
    interfrag_hbonds=True,
    hbond_angles=False,
    freeze_atoms=None,
    define_for=None,
    internals_with_frozen=False,
    rm_for_frag: Optional[set] = None,
    logger=None,
):
    if define_prims is None:
        define_prims = list()
    if freeze_atoms is None:
        freeze_atoms = list()
    if define_for is None:
        define_for = list()
    if rm_for_frag is None:
        rm_for_frag = set()

    log(
        logger,
        f"Detecting primitive internals for {len(atoms)} atoms.\n"
        f"Excluding {len(freeze_atoms)} frozen atoms from the internal coordinate setup.",
    )

    # Mask array. By default all atomes are used to generate internal coordinates.
    use_atoms = np.ones_like(atoms, dtype=bool)
    # Only use atoms in 'define_for' to generate internal coordinates
    if define_for:
        use_atoms[:] = False  # Disable/mask all others
        use_atoms[define_for] = True
    # If not explicitly enabled, don't form internal coordinates containing frozen atoms.
    # With 'internals_with_frozen', the bonds will be filtered for bonds, containing
    # at most one frozen atom.
    elif not internals_with_frozen:
        use_atoms[freeze_atoms] = False
    freeze_atom_set = set(freeze_atoms)
    atoms = [atom for mobile, atom in zip(use_atoms, atoms) if mobile]
    coords3d = coords3d[use_atoms]

    # Maps (different) indices of mobile atoms back to their original indices
    freeze_map = {
        sub_ind: org_ind for sub_ind, org_ind in enumerate(np.where(use_atoms)[0])
    }
    mobile_org_inds = set(freeze_map.values())

    def keep_coord(prim_cls, prim_inds):
        return (
            True
            if (min_weight is None)
            else (prim_cls._weight(atoms, coords3d, prim_inds, 0.12) >= min_weight)
        )

    def keep_coords(prims, prim_cls):
        return [prim for prim in prims if keep_coord(prim_cls, prim)]

    # Bonds
    bonds, cdm, cbm = get_bond_sets(
        atoms,
        coords3d,
        bond_factor=factor,
        return_cdm=True,
        return_cbm=True,
    )
    if internals_with_frozen:
        bonds = [bond for bond in bonds if len(set(bond) & freeze_atom_set) <= 1]
    bonds = [tuple(bond) for bond in bonds]
    bonds = keep_coords(bonds, Stretch)
    bonds = [bond for bond in bonds if rm_for_frag.isdisjoint(set(bond))]

    # Fragments
    fragments = merge_sets(bonds) + [
        frozenset((rmed_atom,)) for rmed_atom in rm_for_frag
    ]
    # Check for unbonded single atoms and create fragments for them.
    bonded_set = set(tuple(np.ravel(bonds)))
    unbonded_set = set(range(len(atoms))) - bonded_set - freeze_atom_set
    log(
        logger,
        f"Merging bonded atoms yielded {len(fragments)} fragment(s) and "
        f"{len(unbonded_set)} atoms.",
    )
    fragments.extend([frozenset((atom,)) for atom in unbonded_set])

    interfrag_bonds = list()
    aux_interfrag_bonds = list()
    translation_inds = list()
    rotation_inds = list()
    # With translational & rotational internal coordinates (TRIC) we don't need
    # interfragment coordinates.
    if tric:
        translation_inds = [list(fragment) for fragment in fragments]
        # Exclude rotational coordinates for atomic species (1 atom)
        rotation_inds = [list(fragment) for fragment in fragments if len(fragment) > 1]
    # Without TRIC we have to somehow connect all fragments.
    else:
        # interfrag_bonds, aux_interfrag_bonds = connect_fragments_kmeans(
        interfrag_bonds, aux_interfrag_bonds = connect_fragments_ahlrichs(
            cdm, fragments, atoms, logger=logger
        )

    # Hydrogen bonds
    assert interfrag_hbonds, "Disabling interfrag_hbonds is not yet supported!"
    hydrogen_bonds = get_hydrogen_bond_inds(atoms, coords3d, bonds, logger=logger)
    hydrogen_set = [frozenset(bond) for bond in hydrogen_bonds]

    def remove_h_bonds(bond_list):
        return [bond for bond in bond_list if set(bond) not in hydrogen_set]

    # Remove newly obtained hydrogen bonds from other lists
    interfrag_bonds = remove_h_bonds(interfrag_bonds)
    aux_interfrag_bonds = remove_h_bonds(aux_interfrag_bonds)
    bonds = remove_h_bonds(bonds)
    aux_bonds = list()  # Not defined by default
    # Don't use auxilary interfragment bonds for bend detection
    bonds_for_bends = [
        bonds,
    ]
    # If we use regular redundant internals (not TRIC) we define interfragment
    # bends.
    if not tric:
        bonds_for_bends += [interfrag_bonds]
        if hbond_angles:
            bonds_for_bends += [hydrogen_bonds]
    bonds_for_bends = set([frozenset(bond) for bond in it.chain(*bonds_for_bends)])

    # Bends
    bends = get_bend_inds(
        coords3d,
        bonds_for_bends,
        min_deg=min_deg,
        # All bends will be checked, for being (close to) linear and will be removed from
        # bend_inds, if needed. Thats why we keep 180° here.
        max_deg=180.0,
        logger=logger,
    )
    bends = keep_coords(bends, Bend)

    # Linear Bends and orthogonal complements
    linear_bends, linear_bend_complements = get_linear_bend_inds(
        coords3d,
        cbm,
        bends,
        min_deg=lb_min_deg,
        max_bonds=lb_max_bonds,
        logger=logger,
    )
    # Remove linear bends from bends
    bends = [bend for bend in bends if bend not in linear_bends]
    linear_bends = keep_coords(linear_bends, LinearBend)
    linear_bend_complements = keep_coords(linear_bend_complements, LinearBend)

    # Dihedrals
    bends_for_dihedrals = bends + linear_bends
    proper_dihedrals, improper_dihedrals = get_dihedral_inds(
        coords3d,
        bonds_for_bends,
        bends_for_dihedrals,
        max_deg=dihed_max_deg,
        logger=logger,
    )
    proper_dihedrals = keep_coords(proper_dihedrals, Torsion)
    improper_dihedrals = keep_coords(improper_dihedrals, Torsion)
    # Improper dihedrals are disabled for now in TRIC
    if tric:
        improper_dihedrals = []

    cartesian_inds = []
    if hybrid:
        cartesian_inds = [i for i, _ in enumerate(atoms)]

    """
    When additional primitives are given in 'define_prims', we want to append
    them to the correct lists, that may contain already some primitive internals
    detected by our algorithms. Here we define a map between the PrimTypes and the
    present lists.
    """
    defined_cartesians = list()
    defined_translations = list()
    defined_rotations = list()

    define_map = {
        PrimTypes.BOND: "bonds",
        PrimTypes.AUX_BOND: "aux_bonds",
        PrimTypes.HYDROGEN_BOND: "hydrogen_bonds",
        PrimTypes.INTERFRAG_BOND: "interfrag_bonds",
        PrimTypes.AUX_INTERFRAG_BOND: "aux_interfrag_bonds",
        PrimTypes.BEND: "bends",
        PrimTypes.LINEAR_BEND: "linear_bends",
        PrimTypes.LINEAR_BEND_COMPLEMENT: "linear_bend_complements",
        PrimTypes.PROPER_DIHEDRAL: "proper_dihedrals",
        PrimTypes.IMPROPER_DIHEDRAL: "improper_dihedrals",
    }
    unmapped_typed_prims = list()
    for type_, *indices in define_prims:
        try:
            key = define_map[type_]
        except KeyError:
            try:
                key = define_map[PrimTypes(type_)]
            except KeyError:
                """
                With the current approach, some primitives in 'define_prims' can't
                be mapped to their respective lists, e.g., given CARTESIAN_X/Y/Z.
                Currently, every item in 'cartesian_inds' is expanded to three
                Cartesians (X/Y/Z). When only the X-component is to be defined, adding
                only the atom index to the list would result in all three components
                to be generated.

                So instead of adding them to their respective lists we keep them in
                'unmapped_typed_prims' and use them later as is.
                """
                unmapped_typed_prims.append((type_, *indices))
                continue
        locals()[key].append(tuple(indices))

    def make_tp(prim_type, *indices):
        """Map possibly modified indices to their original indices.

        With frozen atoms, the indices used to set up internal coordinates do
        not correspond to the actual indices. Here we map them back.
        """
        try:
            org_indices = [freeze_map[ind] for ind in indices]
            """The given 'indices' may not be present in freeze_map. This can happen,
            when coordinates are rebuilt, frozen atoms are excluded from using them
            in the coordinate definition and a previous set of coordinates ALREADY
            using the original indices is supplied."""
        except KeyError as error:
            """In such a case we check if the given coordinates are already fully
            defined in terms of original indices. If so, we use them as is."""
            if set(indices) < mobile_org_inds:
                org_indices = indices
            else:
                raise error
        return (prim_type, *org_indices)

    # Shortcut for PrimTypes Enum
    pt = PrimTypes
    # Create actual typed prims with the desired indices
    typed_prims = (
        # Bonds, two indices
        [make_tp(pt.BOND, *bond) for bond in bonds]
        + [make_tp(pt.AUX_BOND, *abond) for abond in aux_bonds]
        + [make_tp(pt.HYDROGEN_BOND, *hbond) for hbond in hydrogen_bonds]
        + [make_tp(pt.INTERFRAG_BOND, *ifbond) for ifbond in interfrag_bonds]
        + [make_tp(pt.AUX_INTERFRAG_BOND, *aifbond) for aifbond in aux_interfrag_bonds]
        # Bends, three indices
        + [make_tp(pt.BEND, *bend) for bend in bends]
        + [make_tp(pt.LINEAR_BEND, *lbend) for lbend in linear_bends]
        + [
            make_tp(pt.LINEAR_BEND_COMPLEMENT, *lbendc)
            for lbendc in linear_bend_complements
        ]
        # Dihedral, four indices
        + [make_tp(pt.PROPER_DIHEDRAL, *pdihedral) for pdihedral in proper_dihedrals]
        + [
            make_tp(pt.IMPROPER_DIHEDRAL, *idihedral)
            for idihedral in improper_dihedrals
        ]
        + [make_tp(pt.CARTESIAN_X, cind) for cind in cartesian_inds]
        + [make_tp(pt.CARTESIAN_Y, cind) for cind in cartesian_inds]
        + [make_tp(pt.CARTESIAN_Z, cind) for cind in cartesian_inds]
        + [make_tp(pt.CARTESIAN_Z, cind) for cind in cartesian_inds]
    )

    # Translational and rotational coordinates result in 3 different coordinates each
    for frag in translation_inds:
        typed_prims += [
            make_tp(pt.TRANSLATION_X, *frag),
            make_tp(pt.TRANSLATION_Y, *frag),
            make_tp(pt.TRANSLATION_Z, *frag),
        ]
    for frag in rotation_inds:
        typed_prims += [
            make_tp(pt.ROTATION_A, *frag),
            make_tp(pt.ROTATION_B, *frag),
            make_tp(pt.ROTATION_C, *frag),
        ]
    typed_prims += unmapped_typed_prims
    # Drop duplicated typed_prims
    typed_prims = tuple(dict.fromkeys(typed_prims))

    coord_info = CoordInfo(
        bonds=bonds,
        hydrogen_bonds=hydrogen_bonds,
        interfrag_bonds=interfrag_bonds,
        aux_interfrag_bonds=aux_interfrag_bonds,
        bends=bends,
        linear_bends=linear_bends,
        linear_bend_complements=linear_bend_complements,
        proper_dihedrals=proper_dihedrals,
        improper_dihedrals=improper_dihedrals,
        translation_inds=translation_inds,
        rotation_inds=rotation_inds,
        cartesian_inds=cartesian_inds,
        typed_prims=typed_prims,
        fragments=fragments,
    )
    return coord_info


def setup_redundant_from_geom(geom, *args, **kwargs):
    return setup_redundant(geom.atoms, geom.coords3d, *args, **kwargs)


def get_primitives(coords3d, typed_prims, logger=None):
    primitives = list()
    for type_, *indices in typed_prims:
        cls = PrimMap[type_]
        cls_kwargs = {"indices": indices}
        if type_ in Rotations:
            cls_kwargs["ref_coords3d"] = coords3d
        primitives.append(cls(**cls_kwargs))

    msg = (
        "Defined primitives\n"
        + "\n".join(
            [f"\t{i:03d}: {str(p.indices): >14}" for i, p in enumerate(primitives)]
        )
        + "\n"
    )
    log(logger, msg)
    return primitives



import sys

import numpy as np

from pysisyphus.Geometry import Geometry
from pysisyphus.interpolate.Interpolator import Interpolator
from pysisyphus.intcoords.exceptions import (
    DifferentPrimitivesException,
    RebuiltInternalsException,
)
from pysisyphus.intcoords.helpers import get_tangent, form_coordinate_union
from pysisyphus.xyzloader import write_geoms_to_trj


class Redund(Interpolator):
    def __init__(self, *args, align=True, rebuild_geoms=True, **kwargs):
        super().__init__(*args, align=align, **kwargs)

        self.rebuild_geoms = rebuild_geoms
        if self.rebuild_geoms:
            self.geoms = [
                Geometry(geom.atoms, geom.cart_coords, coord_type="redund")
                for geom in self.geoms
            ]

    def dump_progress(self, geoms, out_fn="redund_interpol_fail_trj.xyz"):
        write_geoms_to_trj(geoms, out_fn)
        print(f"Dumped interpolation progress to '{out_fn}'.")

    def interpolate(
        self,
        initial_geom,
        final_geom,
        interpolate_only=0,
        extrapolate=False,
        typed_prims=None,
    ):
        print(f"No. of primitives at initial structure: {initial_geom.coords.size}")
        print(f"No. of primitives at final structure: {final_geom.coords.size}")

        if typed_prims is None:
            typed_prims = form_coordinate_union(initial_geom, final_geom)
            print(f"Union of primitives: {len(typed_prims)}")
        else:
            print(f"Using supplied primitive internals ({len(typed_prims)}).")

        # Recreate geometries with consistent set of internal coordinates
        geom1 = Geometry(
            initial_geom.atoms,
            initial_geom.cart_coords,
            coord_type="redund",
            coord_kwargs={
                "typed_prims": typed_prims,
                "recalc_B": True,
            },
        )
        geom2 = Geometry(
            final_geom.atoms,
            final_geom.cart_coords,
            coord_type="redund",
            coord_kwargs={
                "typed_prims": typed_prims,
                "recalc_B": True,
            },
        )

        dihedral_indices = geom1.internal.dihedral_indices
        initial_tangent = get_tangent(geom1.coords, geom2.coords, dihedral_indices)
        initial_diff = np.linalg.norm(initial_tangent)
        approx_step_size = initial_diff / (self.between + 1)
        final_prims = geom2.internal.prim_coords

        geoms = [
            geom1,
        ]

        def restart(new_geom):
            interpolate_kwargs = {
                "initial_geom": initial_geom,
                "final_geom": final_geom,
                "extrapolate": extrapolate,
                "interpolate_only": interpolate_only,
            }
            return self.restart_interpolate(geoms, new_geom, interpolate_kwargs)

        # initial_geom, final_geom, geoms, new_geom)

        interpolations = interpolate_only if interpolate_only else self.between
        ip_ex = "Extrapolating" if extrapolate else "Interpolating"
        for i in range(interpolations):
            print(f"{ip_ex} {i+1:03d}/{interpolations:03d}")
            new_geom = geoms[-1].copy()
            # Try to use the target primtive internals (final_prims) to calculate
            # a tangent at the current, new geometry. As some primitives may be
            # invalid at 'new_geom', DifferentPrimitivesException may be raised.
            try:
                prim_tangent = get_tangent(
                    new_geom.coords, final_prims, dihedral_indices
                )
            # If this is raised we update our target primitives and try to continue
            # with them.
            except DifferentPrimitivesException:
                # return self.restart_interpolate(initial_geom, final_geom, new_geom)
                return restart(new_geom)
            if extrapolate:
                prim_tangent *= -1
                approx_step_size *= self.extrapolate_damp

            step = self.step_along_tangent(new_geom, prim_tangent, approx_step_size)
            try:
                new_coords = new_geom.coords + step
            # Will be raised when the sizes of both vectors differ
            except ValueError:
                # return self.restart_interpolate(initial_geom, final_geom, new_geom)
                return restart(new_geom)

            # The current primitives may also break down when a step along the
            # tangent is taken.
            try:
                new_geom.coords = new_coords
            except RebuiltInternalsException:
                # return self.restart_interpolate(initial_geom, final_geom, new_geom)
                return restart(new_geom)

            geoms.append(new_geom)
            sys.stdout.flush()

        print()
        return geoms[1:]

    # def restart_interpolate(self, initial_geom, final_geom, geoms, new_geom):
    def restart_interpolate(self, geoms, new_geom, interpolate_kwargs):
        new_typed_prims = new_geom.internal.typed_prims
        print(
            f"Encountered breakdown of current primitive internals.\n"
            "Restarting interpolation with reduced number of internals."
        )
        self.dump_progress(geoms, out_fn=f"redund_interpol_fail_trj.xyz")
        print()
        # Recursive call with reduced set of primitive internals
        return self.interpolate(**interpolate_kwargs, typed_prims=new_typed_prims)

    def step_along_tangent(self, geom, prim_tangent, step_size):
        # Form active set
        B = geom.internal.B_prim
        G = B.dot(B.T)
        eigvals, eigvectors = np.linalg.eigh(G)
        U = eigvectors[:, np.abs(eigvals) > 1e-6]
        reduced_tangent = (np.einsum("i,ij->j", prim_tangent, U) * U).sum(axis=1)
        reduced_tangent /= np.linalg.norm(reduced_tangent)
        step = step_size * reduced_tangent
        return step



from typing import Literal, Union

import numpy as np

from enum import IntEnum


L_MAP = {
    "s": 0,
    "p": 1,
    "d": 2,
    "f": 3,
    "g": 4,
    "h": 5,
}
L_MAP_INV = {key: L for L, key in L_MAP.items()}
L_SIZE = {l: (l + 1) * (l + 2) // 2 for l in L_MAP.values()}

Ls = Literal["s", "p", "d", "f", "g", "h"]
L_Inp = Union[int, Ls]


def get_l(l_inp: L_Inp) -> int:
    """Convert shell label to angular moment quantum number l."""
    try:
        l = L_MAP[l_inp.lower()]
    except (KeyError, AttributeError):
        l = int(l_inp)
    return l


def get_shell_shape(*Ls):
    return [L_SIZE[L] for L in Ls]


class BFType(IntEnum):
    CARTESIAN = 1
    PURE_SPHERICAL = 2


def canonical_order(L):
    inds = list()
    for i in range(L + 1):
        l = L - i
        for n in range(i + 1):
            m = i - n
            inds.append((l, m, n))
    return inds


def cca_order(l):
    """Same as canonical_order()."""
    inds = [
        (l, 0, 0),
    ]
    a = l
    b = c = 0
    for _ in range(((l + 1) * (l + 2) // 2) - 1):
        if c < l - a:
            b -= 1
            c += 1
        else:
            a -= 1
            c = 0
            b = l - a
        inds.append((a, b, c))
    return inds


def cart2ang_moms(cart):
    """Translate 'xxx' to (3, 0, 0) etc."""

    ang_moms = np.zeros(3, dtype=int)
    for char in cart:
        # x: 120, y: 121, z: 122
        ang_moms[ord(char.lower()) - 120] += 1
    return tuple(ang_moms)


def permut_for_order(cart_orders, pysis_order_func=cca_order):
    """Create permutation matrices to reorder Cartesian integrals."""
    Ps = dict()
    for cart_order in cart_orders:
        lengths = [len(_) for _ in cart_order]
        l0 = lengths[0]
        assert all([length == l0 for length in lengths])
        other_order = [cart2ang_moms(_) for _ in cart_order]
        pysis_order = pysis_order_func(l0)

        num = len(pysis_order)
        P = np.zeros((num, num), dtype=int)
        for i, po in enumerate(other_order):
            ind = pysis_order.index(po)
            P[i, ind] = 1
        Ps[l0] = P
    return Ps


def symmetric_orthogonalization(mat, S, thresh=1e-6):
    w, v = np.linalg.eigh(mat.T.dot(S).dot(mat))
    mask = w >= thresh
    w_isqrt = np.sqrt(w[mask])
    S_isqrt = (v[:, mask] / w_isqrt).dot(v[:, mask].T)
    return mat.dot(S_isqrt)



# pdb2reaction/trj2fig.py

"""
Energy-profile plotting utility for XYZ trajectories.

Example:
    pdb2reaction trj2fig -i traj.xyz --reverse-x

For detailed documentation, see: docs/trj2fig.md
"""

from __future__ import annotations

import argparse
import csv
import logging
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import click
import plotly.graph_objs as go
from ase import Atoms
from ase.io import read
from pysisyphus.constants import AU2KCALPERMOL, ANG2BOHR

from .utils import read_xyz_energies

logger = logging.getLogger(__name__)

AXIS_WIDTH = 3         # axis and tick thickness
FONT_SIZE = 18         # tick-label font size
AXIS_TITLE_SIZE = 20   # axis-title font size
LINE_WIDTH = 2         # curve width
MARKER_SIZE = 6        # marker size


# ---------------------------------------------------------------------
#  File helpers
# ---------------------------------------------------------------------
def read_energies_xyz(fname: Path | str) -> List[float]:
    """Compatibility wrapper for utils.read_xyz_energies()."""
    return read_xyz_energies(fname)


def recompute_energies(
    traj_path: Path, charge: Optional[int], multiplicity: Optional[int],
    backend: str = "uma", solvent: str = "none", solvent_model: str = "alpb",
) -> List[float]:
    """
    Recalculate Hartree energies for every frame using the backend factory.
    """
    # Import lazily so comment-only mode does not require torch/UMA deps.
    from .backends import create_calculator

    frames_obj = read(traj_path, index=":", format="xyz")
    frames = [frames_obj] if isinstance(frames_obj, Atoms) else list(frames_obj)
    if not frames:
        raise RuntimeError(f"No frames found in {traj_path}")

    calc = create_calculator(
        backend=backend, charge=charge or 0, spin=multiplicity or 1,
        solvent=solvent, solvent_model=solvent_model,
    )
    energies: List[float] = []
    for atoms in frames:
        elems = atoms.get_chemical_symbols()
        coords_bohr = atoms.get_positions() * ANG2BOHR
        energies.append(float(calc.get_energy(elems, coords_bohr)["energy"]))

    return energies


# ---------------------------------------------------------------------
#  Transformation
# ---------------------------------------------------------------------
def _parse_reference_spec(spec: str | None) -> str | int | None:
    """
    Normalize the reference specification:
      - 'init' (case-insensitive) -> 'init'
      - 'none'/'null'             -> None
      - integer-like string       -> int
    """
    if spec is None:
        return "init"
    s = str(spec).strip()
    lower = s.lower()
    if lower in {"none", "null"}:
        return None
    if lower == "init":
        return "init"
    try:
        return int(s)
    except ValueError:
        raise ValueError(
            f"Invalid -r/--reference: {spec!r}. Use 'init', 'None', or an integer index."
        )


def _resolve_reference_index(
    n_frames: int, ref_spec: str | int | None, reverse_x: bool
) -> Tuple[Optional[int], bool]:
    """
    Decide the reference index and whether to compute a ΔE series.

    Returns (reference_index or None, is_delta).
    """
    if ref_spec is None:
        return None, False  # absolute energies
    if ref_spec == "init":
        idx = 0 if not reverse_x else n_frames - 1
        return idx, True
    idx = int(ref_spec)
    if idx < 0 or idx >= n_frames:
        raise IndexError(f"Reference index {idx} out of range (0..{n_frames-1}).")
    return idx, True


def transform_series(
    energies_hartree: Sequence[float],
    ref_spec_raw: str | None,
    unit: str,
    reverse_x: bool,
) -> Tuple[List[float], str, bool]:
    """
    Compute the y-series and its axis label.

    Returns (values, ylabel, is_delta).
    """
    ref_spec = _parse_reference_spec(ref_spec_raw)
    ref_idx, is_delta = _resolve_reference_index(len(energies_hartree), ref_spec, reverse_x)

    scale = AU2KCALPERMOL if unit == "kcal" else 1.0
    if is_delta:
        base = energies_hartree[ref_idx]  # type: ignore[index]
        values = [float((e - base) * scale) for e in energies_hartree]
        ylabel = f"ΔE ({'kcal/mol' if unit == 'kcal' else 'hartree'})"
    else:
        values = [float(e * scale) for e in energies_hartree]
        ylabel = f"E ({'kcal/mol' if unit == 'kcal' else 'hartree'})"

    return values, ylabel, is_delta


# ---------------------------------------------------------------------
#  Plotting
# ---------------------------------------------------------------------
def _axis_template() -> dict:
    return dict(
        showline=True,
        linewidth=AXIS_WIDTH,
        linecolor="#1C1C1C",
        mirror=True,
        ticks="inside",
        tickwidth=AXIS_WIDTH,
        tickcolor="#1C1C1C",
        tickfont=dict(size=FONT_SIZE, color="#1C1C1C"),
        gridcolor="lightgrey",
        gridwidth=0.5,
        zeroline=False,
    )


def build_figure(delta_or_abs: Sequence[float], ylabel: str, reverse_x: bool) -> go.Figure:
    """
    Build a Plotly figure without a title.
    """
    fig = go.Figure(
        go.Scatter(
            x=list(range(len(delta_or_abs))))
    )
    fig.data[0].update(
        y=list(delta_or_abs),
        mode="lines+markers",
        marker=dict(size=MARKER_SIZE),
        line=dict(shape="spline", smoothing=1.0, width=LINE_WIDTH),
    )

    xaxis_conf = _axis_template() | {
        "title": dict(text="Frame", font=dict(size=AXIS_TITLE_SIZE, color="#1C1C1C"))
    }
    if reverse_x:
        xaxis_conf["autorange"] = "reversed"

    fig.update_layout(
        xaxis=xaxis_conf,
        yaxis=_axis_template() | {
            "title": dict(text=ylabel, font=dict(size=AXIS_TITLE_SIZE, color="#1C1C1C"))
        },
        plot_bgcolor="white",
        paper_bgcolor="white",
        margin=dict(l=80, r=40, t=40, b=80),
    )
    return fig


def save_outputs(
    outs: Sequence[Path],
    fig: Optional[go.Figure],
    energies: Sequence[float],
    values: Sequence[float],
    unit: str,
    is_delta: bool,
) -> None:
    """
    Write all requested outputs.
    """
    for out in outs:
        ext = out.suffix.lower()
        if ext == ".csv":
            write_csv(out, energies, values, unit, is_delta)
        elif ext == ".html":
            assert fig is not None
            fig.write_html(out)
            click.echo(f"[trj2fig] Saved figure -> {out}")
        elif ext in {".png", ".jpg", ".jpeg", ".pdf", ".svg"}:
            assert fig is not None
            kw = {"engine": "kaleido"}
            if ext == ".png":
                kw["scale"] = 2  # high-resolution PNG
            fig.write_image(out, **kw)
            click.echo(f"[trj2fig] Saved figure -> {out}")
        else:
            raise ValueError(f"Unsupported format: {ext}")


def write_csv(
    out: Path,
    energies_hartree: Sequence[float],
    series: Sequence[float],
    unit: str,
    is_delta: bool,
) -> None:
    """
    Save energies (hartree) and ΔE/E series to CSV.
    """
    colname = (f"delta_{unit}" if is_delta else f"energy_{unit}")
    with out.open("w", newline="", encoding="utf-8") as fh:
        w = csv.writer(fh)
        w.writerow(["frame", "energy_hartree", colname])
        for i, (eh, y) in enumerate(zip(energies_hartree, series)):
            w.writerow([i, f"{eh:.8f}", f"{y:.6f}"])
    click.echo(f"[trj2fig] Saved CSV -> {out}")


# ---------------------------------------------------------------------
#  CLI (argparse)
# ---------------------------------------------------------------------
def parse_cli() -> argparse.Namespace:
    p = argparse.ArgumentParser(
        prog="trj2fig",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Plot ΔE or E from an XYZ trajectory and export a figure and/or CSV (no title).",
    )
    p.add_argument("-i", "--input", required=True, help="XYZ trajectory file.")
    p.add_argument(
        "-o",
        "--out",
        nargs="+",
        default=["energy.png"],
        help="Output file(s) [.png/.jpg/.jpeg/.html/.svg/.pdf/.csv]. Multiple names allowed.",
    )
    p.add_argument("--unit", choices=["kcal", "hartree"], default="kcal", help="Energy unit.")
    p.add_argument(
        "-r",
        "--reference",
        default="init",
        help="Reference: 'init' (initial frame; last frame if --reverse-x), 'None' (absolute E), or an integer index.",
    )
    p.add_argument("-q", "--charge", type=int, required=False, help="Total charge. Triggers energy recomputation when supplied.")
    p.add_argument(
        "-m",
        "--multiplicity",
        type=int,
        required=False,
        help="Spin multiplicity (2S+1). Recompute energies when provided.",
    )
    p.add_argument(
        "--reverse-x",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Reverse the x-axis (last frame on the left).",
    )
    return p.parse_args()


def run_trj2fig(
    input_path: Path,
    outs: Sequence[Path],
    unit: str,
    reference: str,
    reverse_x: bool,
    charge: Optional[int] = None,
    multiplicity: Optional[int] = None,
    backend: str = "uma",
    solvent: str = "none",
    solvent_model: str = "alpb",
) -> None:
    traj = input_path.expanduser().resolve()
    if not traj.is_file():
        raise FileNotFoundError(traj)

    if charge is None and multiplicity is None:
        energies = read_energies_xyz(traj)
    else:
        energies = recompute_energies(
            traj, charge, multiplicity,
            backend=backend, solvent=solvent, solvent_model=solvent_model,
        )
    values, ylabel, is_delta = transform_series(energies, reference, unit, reverse_x)

    need_plot = any(Path(o).suffix.lower() != ".csv" for o in outs)
    fig = build_figure(values, ylabel, reverse_x) if need_plot else None

    out_paths = [Path(o).expanduser().resolve() for o in outs]
    save_outputs(out_paths, fig, energies, values, unit, is_delta)


def main() -> None:
    args = parse_cli()
    run_trj2fig(
        Path(args.input),
        args.out,
        args.unit,
        args.reference,
        args.reverse_x,
        args.charge,
        args.multiplicity,
    )


# ---------------------------------------------------------------------
#  Click wrapper for package CLI integration
# ---------------------------------------------------------------------
@click.command(
    name="trj2fig",
    help="Plot ΔE or E from an XYZ trajectory and export figure/CSV.",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.option(
    "-i",
    "--input",
    "input_path",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="XYZ trajectory file.",
)
@click.option(
    "-o",
    "--out",
    "outs",
    multiple=True,
    default=(),
    type=click.Path(dir_okay=False, path_type=Path),
    help="Output file(s). You can repeat -o and/or list extra filenames after options "
         "(.png/.jpg/.jpeg/.html/.svg/.pdf/.csv). If nothing is given, defaults to energy.png.",
)
@click.argument(
    "extra_outs",
    nargs=-1,
    type=click.Path(dir_okay=False, path_type=Path),
)
@click.option(
    "--unit",
    type=click.Choice(["kcal", "hartree"]),
    default="kcal",
    help="Energy unit.",
)
@click.option(
    "-r",
    "--reference",
    default="init",
    help="Reference: 'init' (initial frame; last frame if --reverse-x), 'None' (absolute E), or an integer index.",
)
@click.option(
    "-q",
    "--charge",
    type=int,
    default=None,
    help="Total charge. Triggers energy recomputation when supplied.",
)
@click.option(
    "-m",
    "--multiplicity",
    type=int,
    default=None,
    help="Spin multiplicity (2S+1). Triggers energy recomputation when supplied.",
)
@click.option(
    "--reverse-x/--no-reverse-x",
    default=False,
    show_default=True,
    help="Reverse the x-axis (last frame on the left).",
)
@click.option("-b", "--backend", type=click.Choice(["uma", "orb", "mace", "aimnet2"]), default="uma",
              show_default=True, help="MLIP backend.")
@click.option("--solvent", default="none", show_default=True,
              help="Implicit solvent name for xTB correction (e.g. 'water'). 'none' to disable.")
@click.option("--solvent-model", "solvent_model", default="alpb", type=click.Choice(["alpb", "cpcmx"]),
              show_default=True, help="xTB solvent model.")
def cli(
    input_path: Path,
    outs: Tuple[Path, ...],
    extra_outs: Tuple[Path, ...],
    unit: str,
    reference: str,
    charge: Optional[int],
    multiplicity: Optional[int],
    reverse_x: bool,
    backend: str,
    solvent: str,
    solvent_model: str,
) -> None:
    # Combine outputs from -o with positional filenames that follow the options
    all_outs: List[Path] = list(outs) + list(extra_outs)
    if not all_outs:
        all_outs = [Path("energy.png")]
    run_trj2fig(input_path, all_outs, unit, reference, reverse_x, charge, multiplicity,
                backend=backend, solvent=solvent, solvent_model=solvent_model)



from math import sqrt

import numpy as np
from scipy.optimize import root_scalar

from pysisyphus.optimizers.HessianOptimizer import HessianOptimizer


class RSA(HessianOptimizer):
    """The Importance of Step Control in Optimization Methods, del Campo, 2009."""

    def optimize(self):
        energy, gradient, H, big_eigvals, big_eigvecs, resetted = self.housekeeping()

        assert big_eigvals.argmin() == 0
        min_eigval = big_eigvals[0]
        pos_definite = min_eigval > 0.0
        gradient_trans = big_eigvecs.T.dot(gradient)
        # This will be also be True when we come close to a minimizer,
        # but then the Hessian will also be positive definite and a
        # simple Newton step will be used.
        hard_case = abs(gradient_trans[0]) <= 1e-6
        self.log(f"Smallest eigenvalue: {min_eigval:.6f}")
        self.log(f"Positive definite Hessian: {pos_definite}")
        self.log(f"Hard case: {hard_case}")

        def get_step(lambda_):
            return -gradient_trans / (big_eigvals + lambda_)

        # Unshifted Newton step
        newton_step = get_step(0.0)
        newton_norm = np.linalg.norm(newton_step)

        # def on_trust_radius(step, thresh=1e-3):
        # return abs(self.trust_radius - np.linalg.norm(step)) <= thresh

        def on_trust_radius_lin(step):
            return 1 / self.trust_radius - 1 / np.linalg.norm(step)

        def finalize_step(lambda_):
            step = get_step(lambda_)
            step = big_eigvecs.dot(step)
            predicted_change = step.dot(gradient) + 0.5 * step.dot(H).dot(step)
            self.predicted_energy_changes.append(predicted_change)
            return step

        # Simplest case. Positive definite Hessian and predicted step is
        # already in trust radius.
        if pos_definite and newton_norm <= self.trust_radius:
            lambda_ = 0.0
            self.log("Using unshifted Newton step.")
            return finalize_step(lambda_)

        # If the Hessian is not positive definite or if the step is too
        # long we have to determine the shift parameter lambda.
        rs_kwargs = {
            "f": lambda lambda_: on_trust_radius_lin(get_step(lambda_)),
            "xtol": 1e-3,
            # Would otherwise be chosen automatically, but we set it
            # here explicitly for verbosity.
            "method": "brentq",
        }

        def root_search(bracket):
            rs_kwargs.update(
                {
                    "bracket": bracket,
                    "x0": bracket[0] + 1e-3,
                }
            )
            res = root_scalar(**rs_kwargs)
            return res

        BRACKET_END = 1e10
        if not hard_case:
            bracket_start = 0.0 if pos_definite else -min_eigval - 1e-3
            bracket = (bracket_start, BRACKET_END)
            res = root_search(bracket)
            assert res.converged
            return finalize_step(res.root)

        # Hard case.
        # First we try the bracket (-b1, ∞)
        bracket = (-min_eigval, BRACKET_END)
        res = root_search(bracket)
        if res.converged:
            return finalize_step(res.root)

        # Now we would try the bracket (-b2, -b1). The resulting step should have
        # a suitable length, but the (shifted) Hessian would have an incorrect
        # eigenvalue spectrum (not positive definite). To solve this we use a
        # different formula to calculate the step.
        without_first = gradient_trans[1:] / (big_eigvals[1:] - min_eigval)
        tau = sqrt(self.trust_radius ** 2 - (without_first ** 2).sum())
        step_trans = [tau] + -without_first.tolist()
        step = big_eigvecs.dot(step_trans)
        predicted_change = step.dot(gradient) + 0.5 * step.dot(H).dot(step)
        self.predicted_energy_changes.append(predicted_change)
        return step



import abc
from dataclasses import dataclass
import functools
import logging
import os
from pathlib import Path
import sys
import textwrap
import time
from typing import Literal, Optional, Tuple

import numpy as np
import yaml

from pysisyphus.cos.ChainOfStates import ChainOfStates
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import (
    check_for_end_sign,
    fit_rigid,
    get_coords_diffs,
    procrustes,
)
from pysisyphus.helpers_pure import highlight_text
from pysisyphus.intcoords.exceptions import RebuiltInternalsException
from pysisyphus.intcoords.helpers import interfragment_distance
from pysisyphus.io.hdf5 import get_h5_group, resize_h5_group
from pysisyphus.optimizers.exceptions import ZeroStepLength
from pysisyphus.TablePrinter import TablePrinter

import torch

def get_data_model(geometry, is_cos, max_cycles):
    try:
        # Attribute is only present in COS classes
        image_num = geometry.max_image_num
        dummy_geom = geometry.images[0]
    except AttributeError:
        image_num = 1
        dummy_geom = geometry

    # Define dataset shapes. As pysisyphus offers growing COS methods where
    # the number of images changes along the optimization we have to define
    # the shapes accordingly by considering the maximum number of images.
    _1d = (max_cycles,)
    _2d = (max_cycles, image_num * dummy_geom.coords.size)
    _image_inds = (max_cycles, image_num)
    # Number of cartesian coordinates is probably different from the number
    # of internal coordinates.
    _2d_cart = (max_cycles, image_num * dummy_geom.cart_coords.size)
    # The dimensionality of energies depends on whether a COS is optimized or
    # not. I know this is probably not the best idea...
    _energy = _1d if (not is_cos) else (max_cycles, geometry.max_image_num)

    data_model = {
        "image_nums": _1d,
        "image_inds": _image_inds,
        "cart_coords": _2d_cart,
        "coords": _2d,
        "energies": _energy,
        "forces": _2d,
        # AFIR related
        "true_energies": _energy,
        "true_forces": _2d_cart,
        "steps": _2d,
        # Convergence related
        "max_forces": _1d,
        "rms_forces": _1d,
        "max_steps": _1d,
        "rms_steps": _1d,
        # Misc
        "cycle_times": _1d,
        "modified_forces": _2d,
        # COS specific
        "tangents": _2d,
    }

    return data_model


CONV_THRESHS = {
    #                max_force, rms_force, max_step, rms_step
    "nwchem_loose": (4.5e-3, 3.0e-3, 5.4e-3, 3.6e-3),
    "gau_loose": (2.5e-3, 1.7e-3, 1.0e-2, 6.7e-3),
    "gau": (4.5e-4, 3.0e-4, 1.8e-3, 1.2e-3),
    "gau_tight": (1.5e-5, 1.0e-5, 6.0e-5, 4.0e-5),
    "gau_vtight": (2.0e-6, 1.0e-6, 6.0e-6, 4.0e-6),
    "baker": (3.0e-4, 2.0e-4, 3.0e-4, 2.0e-4),
    # Dummy thresholds
    "never": (2.0e-6, 1.0e-6, 6.0e-6, 4.0e-6),
}
Thresh = Literal["gau_loose", "gau", "gau_tight", "gau_vtight", "baker", "never"]


@dataclass(frozen=True)
class ConvInfo:
    cur_cycle: int
    energy_converged: bool
    max_force_converged: bool
    rms_force_converged: bool
    max_step_converged: bool
    rms_step_converged: bool
    desired_eigval_structure: bool

    def get_convergence(self):
        return (
            self.energy_converged,
            self.max_force_converged,
            self.rms_force_converged,
            self.max_step_converged,
            self.rms_step_converged,
            self.desired_eigval_structure,
        )

    def is_converged(self):
        return all(self.get_convergence())


class Optimizer(metaclass=abc.ABCMeta):
    def __init__(
        self,
        geometry: Geometry,
        thresh: Thresh = "gau_loose",
        max_step: float = 0.04,
        max_cycles: int = 150,
        min_step_norm: float = 1e-8,
        assert_min_step: bool = True,
        rms_force: Optional[float] = None,
        rms_force_only: bool = False,
        max_force_only: bool = False,
        force_only: bool = False,
        converge_to_geom_rms_thresh: float = 0.05,
        align: bool = False,
        align_factor: float = 1.0,
        dump: bool = False,
        dump_restart: bool = False,
        print_every: int = 1,
        prefix: str = "",
        reparam_thresh: float = 1e-3,
        reparam_check_rms: bool = True,
        reparam_when: Optional[Literal["before", "after"]] = "after",
        overachieve_factor: float = 0.0,
        check_eigval_structure: bool = False,
        restart_info=None,
        check_coord_diffs: bool = True,
        coord_diff_thresh: float = 0.01,
        fragments: Optional[Tuple] = None,
        monitor_frag_dists: int = 0,
        out_dir: str = ".",
        h5_fn: str = "optimization.h5",
        h5_group_name: str = "opt",
    ) -> None:
        """Optimizer baseclass. Meant to be subclassed.

        Parameters
        ----------
        geometry
            Geometry to be optimized.
        thresh
            Convergence threshold.
        max_step
            Maximum absolute component of the allowed step vector. Utilized in
            optimizers that don't support a trust region or line search.
        max_cycles
            Maximum number of allowed optimization cycles.
        min_step_norm
            Minimum norm of an allowed step. If the step norm drops below
            this value a ZeroStepLength-exception is raised. The unit depends
            on the coordinate system of the supplied geometry.
        assert_min_step
            Flag that controls whether the norm of the proposed step is check
            for being too small.
        rms_force
            Root-mean-square of the force from which user-defined thresholds
            are derived. When 'rms_force' is given 'thresh' is ignored.
        rms_force_only
            When set, convergence is signalled only based on rms(forces).
        max_force_only
            When set, convergence is signalled only based on max(|forces|).
        force_only
            When set, convergence is signalled only based on max(|forces|) and rms(forces).
        converge_to_geom_rms_thresh
            Threshold for the RMSD with another geometry. When the RMSD drops
            below this threshold convergence is signalled. Only used with
            Growing Newton trajectories.
        align
            Flag that controls whether the geometry is aligned in every step
            onto the coordinates of the previous step. Must not be used with
            internal coordinates.
        align_factor
            Factor that controls the strength of the alignment. 1.0 means
            full alignment, 0.0 means no alignment. The factor mixes the
            rotation matrix of the alignment with the identity matrix.
        dump
            Flag to control dumping/writing of optimization progress to the
            filesystem
        dump_restart
            Flag to control whether restart information is dumped to the
            filesystem.
        print_every
            Report optimization progress every nth cycle.
        prefix
            Short string that is prepended to several files created by
            the optimizer. Allows distinguishing several optimizations carried
            out in the same directory.
        reparam_thresh
            Controls the minimal allowed similarity between coordinates
            after two successive reparametrizations. Convergence is signalled
            if the coordinates did not change significantly.
        reparam_check_rms
            Whether to check for (too) similar coordinates after reparametrization.
        reparam_when
            Reparametrize before or after calculating the step. Can also be turned
            off by setting it to None.
        overachieve_factor
            Signal convergence when max(forces) and rms(forces) fall below the
            chosen threshold, divided by this factor. Convergence of max(step) and
            rms(step) is ignored.
        check_eigval_structure
            Check the eigenvalues of the modes we maximize along. Convergence requires
            them to be negative. Useful if TS searches are started from geometries close
            to a minimum.
        restart_info
            Restart information. Undocumented.
        check_coord_diffs
            Whether coordinates of chain-of-sates images are checked for being
            too similar.
        coord_diff_thresh
            Unitless threshold for similary checking of COS image coordinates.
            The first image is assigned 0, the last image is assigned to 1.
        fragments
            Tuple of lists containing atom indices, defining two fragments.
        monitor_frag_dists
            Monitor fragment distances for N cycles. The optimization is terminated
            when the interfragment distances falls below the initial value after N
            cycles.
        out_dir
            String poiting to a directory where optimization progress is
            dumped.
        h5_fn
            Basename of the HDF5 file used for dumping.
        h5_group_name
            Groupname used for dumping of this optimization.
        """
        assert thresh in CONV_THRESHS.keys()

        self.geometry = geometry
        self.thresh = thresh
        self.max_step = max_step
        self.min_step_norm = min_step_norm
        self.assert_min_step = assert_min_step
        self.rms_force_only = rms_force_only
        self.max_force_only = max_force_only
        self.force_only = force_only
        self.converge_to_geom_rms_thresh = converge_to_geom_rms_thresh
        self.align = align
        self.align_factor = align_factor
        self.dump = dump
        self.dump_restart = dump_restart
        print_every = int(print_every)
        assert print_every >= 1
        self.print_every = print_every
        self.prefix = f"{prefix}_" if prefix else prefix
        self.reparam_thresh = reparam_thresh
        self.reparam_check_rms = reparam_check_rms
        self.reparam_when = reparam_when
        assert self.reparam_when in ("after", "before", None)
        self.overachieve_factor = float(overachieve_factor)
        self.check_eigval_structure = check_eigval_structure
        self.check_coord_diffs = check_coord_diffs
        self.coord_diff_thresh = float(coord_diff_thresh)

        self.logger = logging.getLogger("optimizer")
        self.is_cos = issubclass(type(self.geometry), ChainOfStates)

        # Set up convergence thresholds
        self.convergence = self.make_conv_dict(
            thresh, rms_force, rms_force_only, max_force_only, force_only
        )
        for key, value in self.convergence.items():
            setattr(self, key, value)

        if self.thresh == "never":
            max_cycles = 1_000_000_000
            self.dump = False
            self.log(
                f"Got threshold {self.thresh}, set 'max_cycles' to {max_cycles} "
                "and disabled dumping!"
            )
            self.conv_dict = {}
        try:
            self.converge_to_geom = self.geometry.converge_to_geom
        except AttributeError:
            # TODO: log that attribute is not present at debug level
            self.converge_to_geom = None
        self.max_cycles = max_cycles

        self.fragments = fragments
        self.monitor_frag_dists = monitor_frag_dists
        if self.monitor_frag_dists:
            assert (
                len(self.fragments) == 2
            ), "Interfragment monitoring requires two fragments!"
            assert all(
                [len(frag) > 0 for frag in self.fragments]
            ), "Fragments must not be empty!"
        # Setting some default values
        self.monitor_frag_dists_counter = self.monitor_frag_dists
        self.interfrag_dists = list()
        self.resetted = False
        try:
            out_dir = Path(out_dir)
        except TypeError:
            out_dir = Path(".")
        self.out_dir = out_dir.resolve()
        self.out_dir.mkdir(parents=True, exist_ok=True)

        if self.is_cos:
            moving_image_num = len(self.geometry.moving_indices)
            print(f"Path with {moving_image_num} moving images.")

        # Don't use prefix for this fn, as different optimizations
        # can be distinguished according to their group in the HDF5 file.
        self.h5_fn = self.get_path_for_fn(h5_fn, with_prefix=False)
        self.h5_group_name = h5_group_name

        current_fn = "current_geometries_trj.xyz" if self.is_cos else "current_geometry.xyz"
        self.current_fn = self.get_path_for_fn(current_fn)
        final_fn = "final_geometries_trj.xyz" if self.is_cos else "final_geometry.xyz"
        self.final_fn = self.get_path_for_fn(final_fn)
        self.hei_trj_fn = self.get_path_for_fn("cos_hei_trj.xyz")
        try:
            os.remove(self.hei_trj_fn)
        except FileNotFoundError:
            pass

        # Setting some empty lists as default. The actual shape of the respective
        # entries is not considered, which gives us some flexibility.
        self.data_model = get_data_model(self.geometry, self.is_cos, self.max_cycles)
        for la in self.data_model.keys():
            setattr(self, la, list())

        if self.dump:
            out_trj_fn = self.get_path_for_fn("optimization_trj.xyz")
            self.out_trj_handle = open(out_trj_fn, "w")
            # Call with reset=True to delete remnants of previous calculations, unless
            # the optimizer was restarted. Given a previous optimization with, e.g. 30
            # cycles and a second restarted optimization with 20 cycles the last 10 cycles
            # of the previous optimization would still be present.
            reset = restart_info is None
            # h5_group = get_h5_group(
            #     self.h5_fn, self.h5_group_name, self.data_model, reset=reset
            # )
            # h5_group.file.close()
        if self.prefix:
            self.log(f"Created optimizer with prefix {self.prefix}")

        self.restarted = False
        self.last_cycle = 0
        self.cur_cycle = 0
        if restart_info is not None:
            if isinstance(restart_info, str):
                restart_info = yaml.load(restart_info, Loader=yaml.SafeLoader)
            self.set_restart_info(restart_info)
            self.restarted = True

        header = "cycle Δ(energy) max(|force|) rms(force) max(|step|) rms(step) s/cycle".split()
        col_fmts = "int float float float float float float_short".split()
        self.table = TablePrinter(header, col_fmts, width=12)
        self.is_converged = False

    def get_path_for_fn(self, fn, with_prefix=True):
        prefix = self.prefix if with_prefix else ""
        return self.out_dir / (prefix + fn)

    def make_conv_dict(
        self, key, rms_force=None, rms_force_only=False, max_force_only=False, force_only=False
    ):
        if not rms_force:
            threshs = CONV_THRESHS[key]
        else:
            print(
                "Deriving convergence threshold from supplied "
                f"rms_force={rms_force}."
            )
            threshs = (
                1.5 * rms_force,
                rms_force,
                6 * rms_force,
                4 * rms_force,
            )
        keys = keep_keys = [
            "max_force_thresh",
            "rms_force_thresh",
            "max_step_thresh",
            "rms_step_thresh",
        ]
        conv_dict = {k: v for k, v in zip(keys, threshs)}

        # Only used gradient information for COS optimizations
        if self.is_cos:
            keep_keys = ["max_force_thresh", "rms_force_thresh"]

        if rms_force_only:
            self.log("Checking convergence with rms(forces) only!")
            keep_keys = ["rms_force_thresh"]
        elif max_force_only:
            self.log("Checking convergence with max(forces) only!")
            keep_keys = ["max_force_thresh"]
        elif force_only:
            self.log("Checking convergence with max(forces) and rms(forces) only!")
            keep_keys = ["max_force_thresh", "rms_force_thresh"]

        # The dictionary should only contain pairs that are needed
        conv_dict = {key: value for key, value in conv_dict.items() if key in keep_keys}
        return conv_dict

    def report_conv_thresholds(self):
        oaf = self.overachieve_factor

        # Overachieved
        def oa(val):
            return f", ({val/oaf:.6f})" if oaf > 0.0 else ""

        internal_coords = self.geometry.coord_type not in (
            "cart",
            "cartesian",
            "mwcartesian",
        )
        fu = "E_h a_0⁻¹" + (" (rad⁻¹)" if internal_coords else "")  # forces unit
        su = "a_0" + (" (rad)" if internal_coords else "")  # step unit

        try:
            rms_thresh = f"\tmax(|force|) <= {self.max_force_thresh:.6f}{oa(self.max_force_thresh)} {fu}"
        except AttributeError:
            rms_thresh = None
        try:
            max_thresh = f"\t  rms(force) <= {self.rms_force_thresh:.6f}{oa(self.rms_force_thresh)} {fu}"
        except AttributeError:
            max_thresh = None
        threshs = (rms_thresh, max_thresh)

        if self.rms_force_only:
            use_threshs = (threshs[1],)
        elif self.max_force_only:
            use_threshs = (threshs[0],)
        elif self.force_only:
            use_threshs = threshs
        elif self.is_cos:
            use_threshs = threshs
        else:
            use_threshs = threshs + (
                f"\t max(|step|) <= {self.max_step_thresh:.6f} {su}",
                f"\t   rms(step) <= {self.rms_step_thresh:.6f} {su}",
            )
        if self.thresh == "baker":
            use_threshs = use_threshs + ("\t   Δ(energy) <= 0.000001 E_h",)
        print(
            "Convergence thresholds"
            + (", (overachieved when)" if oaf > 0.0 else "")
            + ":\n"
            + "\n".join(use_threshs)
            + f"\n\t'Superscript {self.table.mark}' indicates convergence"
            + "\n"
        )

    def log(self, message, level=50):
        self.logger.log(level, message)

    def check_convergence(self, step=None, multiple=1.0, overachieve_factor=None):
        """Check if the current convergence of the optimization
        is equal to or below the required thresholds, or a multiple
        thereof. The latter may be used in initiating the climbing image.
        """

        if step is None:
            step = self.steps[-1]
        if overachieve_factor is None:
            overachieve_factor = self.overachieve_factor

        if isinstance(step, torch.Tensor):
            step = step.detach().cpu().numpy()

        # When using a ChainOfStates method we are only interested
        # in optimizing the forces perpendicular to the MEP.
        if self.is_cos:
            forces = self.geometry.perpendicular_forces
        elif len(self.modified_forces) == len(self.forces):
            self.log("Using modified forces to determine convergence!")
            forces = self.modified_forces[-1]
        else:
            forces = self.forces[-1]

        # The forces of fixed images may be zero and this may distort the RMS
        # values. So we take into account the number of moving images with
        # non-zero forces vectors.
        if self.is_cos:
            non_zero_elements = (
                len(self.geometry.moving_indices) * self.geometry.coords_length
            )
            rms_force = np.sqrt(np.sum(np.square(forces)) / non_zero_elements)
            rms_step = np.sqrt(np.sum(np.square(step)) / non_zero_elements)
        else:
            rms_force = np.sqrt(np.mean(np.square(forces)))
            rms_step = np.sqrt(np.mean(np.square(step)))

        max_force = np.abs(forces).max()
        max_step = np.abs(step).max()

        self.max_forces.append(max_force)
        self.rms_forces.append(rms_force)
        self.max_steps.append(max_step)
        self.rms_steps.append(rms_step)

        # Give geometry a chance to signal convergence, e.g. GrowingNT that
        # is supposed to stop when a TS was passed.
        try:
            geom_converged = self.geometry.check_convergence()
        except AttributeError:
            geom_converged = False

        converged_to_geom = False
        if self.converge_to_geom is not None:
            rmsd = np.sqrt(
                np.mean((self.converge_to_geom.coords - self.geometry.coords) ** 2)
            )
            converged_to_geom = rmsd < self.converge_to_geom_rms_thresh

        this_cycle = {
            "max_force_thresh": max_force,
            "rms_force_thresh": rms_force,
            "max_step_thresh": max_step,
            "rms_step_thresh": rms_step,
        }

        def check(key):
            # Always return True if given key is not checked
            key_is_checked = key in self.convergence
            if key_is_checked:
                result = this_cycle[key] <= getattr(self, key) * multiple
            else:
                result = True
            return result

        convergence = {
            "energy_converged": True,
            "max_force_converged": check("max_force_thresh"),
            "rms_force_converged": check("rms_force_thresh"),
            "max_step_converged": check("max_step_thresh"),
            "rms_step_converged": check("rms_step_thresh"),
        }
        # For TS optimizations we also try to check the eigenvalue structure of the
        # Hessian. A saddle point of order n must have exatcly only n significant negative
        # eigenvalues. We try to check this below.
        #
        # Currently, this is not totally strict,
        # as only the values in self.ts_mode_eigvals are checked but actually all eigenvalues
        # would have to be checked.
        desired_eigval_structure = True
        if self.check_eigval_structure:
            try:
                desired_eigval_structure = (
                    # Acutally all eigenvalues would have to be checked, but
                    # currently they are not stored anywhere.
                    self.ts_mode_eigvals
                    < self.small_eigval_thresh
                ).sum() == len(self.roots)
            except AttributeError:
                self.log(
                    "Skipping check of eigenvalue structure, as information is unavailable."
                )
        convergence["desired_eigval_structure"] = desired_eigval_structure
        conv_info = ConvInfo(self.cur_cycle, **convergence)

        # Check if force convergence is overachieved. If the eigenvalue-structure is
        # checked, a wrong structure will prevent overachieved convergence.
        overachieved = False
        if overachieve_factor > 0 and desired_eigval_structure:
            max_thresh = self.convergence.get("max_force_thresh", 0) / overachieve_factor
            rms_thresh = self.convergence.get("rms_force_thresh", 0) / overachieve_factor
            max_ = max_force < max_thresh
            rms_ = rms_force < rms_thresh
            overachieved = max_ and rms_
            if max_:
                self.log("max(force) is overachieved")
            if rms_:
                self.log("rms(force) is overachieved")
            if max_ and rms_:
                self.log("Force convergence overachieved!")

        converged = all([val for val in convergence.values()])
        not_never = self.thresh != "never"

        if self.thresh == "baker":
            energy_converged = False
            if len(self.energies) >= 2:
                cur_energy = np.asarray(self.energies[-1])
                prev_energy = np.asarray(self.energies[-2])
                if cur_energy.shape == prev_energy.shape:
                    energy_converged = np.all(np.abs(cur_energy - prev_energy) < 1e-6)

            # Enforce at least one new cycle after (re)start and require energy convergence.
            convergence["energy_converged"] = bool(energy_converged)
            conv_info = ConvInfo(self.cur_cycle, **convergence)
            converged = (self.cur_cycle > self.last_cycle) and all(convergence.values())
            # Keep Baker strict: don't bypass the energy criterion via overachievement.
            overachieved = False
        return (
            any((converged_to_geom, converged, overachieved, geom_converged))
            and not_never,
            conv_info,
        )

    def print_opt_progress(self, conv_info):
        try:
            energy_diff = self.energies[-1] - self.energies[-2]
        # ValueError: maybe raised when the number of images differ in cycles
        # IndexError: raised in first cycle when only one energy is present
        except (ValueError, IndexError):
            energy_diff = float("nan")

        # Try to sum COS energies
        try:
            energy_diff = sum(energy_diff)
        except TypeError:
            pass

        if (self.cur_cycle > 1) and (self.cur_cycle % 10 == 0):
            self.table.print_sep()

        # desired_eigval_structure is also returned, but currently not reported.
        marks = [False, *conv_info.get_convergence()[:-1], False]
        try:
            cycle_time = self.cycle_times[-1]
        except IndexError:
            cycle_time = 0.0
        self.table.print_row(
            (
                self.cur_cycle,
                energy_diff,
                self.max_forces[-1],
                self.rms_forces[-1],
                self.max_steps[-1],
                self.rms_steps[-1],
                cycle_time,
            ),
            marks=marks,
        )
        try:
            # Geometries/ChainOfStates objects can also do some printing.
            add_info = self.geometry.get_additional_print()
            if add_info:
                self.table.print(add_info)
        except AttributeError:
            pass

    def fit_rigid(self, *, vectors=None, vector_lists=None, hessian=None):
        return fit_rigid(
            self.geometry,
            vectors=vectors,
            vector_lists=vector_lists,
            hessian=hessian,
            align_factor=self.align_factor,
        )

    def procrustes(self):
        """Wrapper for procrustes that passes additional arguments along."""
        procrustes(self.geometry, self.align_factor)

    def scale_by_max_step(self, steps):
        steps_max = np.abs(steps).max()
        if steps_max > self.max_step:
            steps *= self.max_step / steps_max
        return steps

    def prepare_opt(self):
        pass

    def postprocess_opt(self):
        pass

    @abc.abstractmethod
    def optimize(self):
        pass

    def write_to_out_dir(self, out_fn, content, mode="w"):
        out_path = self.out_dir / out_fn
        with open(out_path, mode) as handle:
            handle.write(content)

    def write_image_trjs(self):
        base_name = "image_{:03d}_trj.xyz"
        for i, image in enumerate(self.geometry.images):
            image_fn = base_name.format(i)
            comment = f"cycle {self.cur_cycle}"
            as_xyz = image.as_xyz(comment)
            self.write_to_out_dir(image_fn, as_xyz + "\n", "a")

    def write_results(self):
        # Save results from the Optimizer to HDF5 file if requested
        # h5_group = get_h5_group(self.h5_fn, self.h5_group_name)

        # Some attributes never change and are only set in the first cycle
        # if self.cur_cycle == 0:
        #     h5_group.attrs["is_cos"] = self.is_cos
        #     try:
        #         atoms = self.geometry.images[0].atoms
        #         coord_size = self.geometry.images[0].coords.size
        #     except AttributeError:
        #         atoms = self.geometry.atoms
        #         coord_size = self.geometry.coords.size
        #     h5_group.attrs["atoms"] = np.bytes_(atoms)
        #     h5_group.attrs["coord_type"] = self.geometry.coord_type
        #     h5_group.attrs["coord_size"] = coord_size
        #     h5_group.attrs["overachieve_factor"] = self.overachieve_factor
        #     for key in (
        #         "max_force_thresh",
        #         "rms_force_thresh",
        #         "max_step_thresh",
        #         "rms_step_thresh",
        #     ):
        #         try:
        #             h5_group.attrs[key] = self.convergence[key]
        #         # Step threshold may not be present
        #         except KeyError:
        #             pass

        # Update changing attributes
        # h5_group.attrs["cur_cycle"] = self.cur_cycle
        # h5_group.attrs["is_converged"] = self.is_converged

        # for key, shape in self.data_model.items():
        #     value = getattr(self, key)
        #     # Don't try to set empty values, e.g. 'tangents' are only present
        #     # for COS methods. 'modified_forces' are only present for NCOptimizer.
        #     if not value:
        #         continue
        #     if len(shape) > 1:
        #         h5_group[key][self.cur_cycle, : len(value[-1])] = value[-1]
        #     else:
        #         h5_group[key][self.cur_cycle] = value[-1]

        # h5_group.file.close()
        pass

    def write_cycle_to_file(self):
        as_xyz_str = self.geometry.as_xyz()

        if self.is_cos:
            out_fn = "cycle_{:03d}_trj.xyz".format(self.cur_cycle)
            self.write_to_out_dir(out_fn, as_xyz_str)
            # Also write separate _trj.xyz files for every image in the cos
            self.write_image_trjs()

            # Dump current HEI
            max_ind = np.argmax(self.energies[-1])
            with open(self.hei_trj_fn, "a") as handle:
                handle.write(self.geometry.images[max_ind].as_xyz() + "\n")

        else:
            # Append to _trj.xyz file
            self.out_trj_handle.write(as_xyz_str + "\n")
            self.out_trj_handle.flush()
        # Dump to HDF5
        # self.write_results()

    def final_summary(self):
        # If the optimization was stopped _forces may not be set, so
        # then we force a calculation if it was not already set.
        _ = self.geometry.forces
        cart_forces = self.geometry.cart_forces
        max_cart_forces = np.abs(cart_forces).max()
        rms_cart_forces = np.sqrt(np.mean(cart_forces**2))
        int_str = ""
        if self.geometry.coord_type not in ("cart", "cartesian", "mwcartesian"):
            int_forces = self.geometry.forces
            max_int_forces = np.abs(int_forces).max()
            rms_int_forces = np.sqrt(np.mean(int_forces**2))
            int_str = f"""
            \tmax(forces, internal): {max_int_forces:.6f} hartree/(bohr,rad)
            \trms(forces, internal): {rms_int_forces:.6f} hartree/(bohr,rad)"""
        energy = self.geometry.energy
        final_summary = f"""
        Final summary:{int_str}
        \tmax(forces,cartesian): {max_cart_forces:.6f} hartree/bohr
        \trms(forces,cartesian): {rms_cart_forces:.6f} hartree/bohr
        \tenergy: {energy:.8f} hartree
        """
        return textwrap.dedent(final_summary.strip())

    _au_header_shown = False

    def run(self):
        if not Optimizer._au_header_shown:
            print("If not specified otherwise, all quantities are given in au.\n")
            Optimizer._au_header_shown = True

        if not self.restarted:
            prep_start_time = time.time()
            self.prepare_opt()
            self.log(f"{self.geometry.coords.size} degrees of freedom.")
            prep_end_time = time.time()
            prep_time = prep_end_time - prep_start_time
            self.report_conv_thresholds()
            print(f"Spent {prep_time:.1f} s preparing the first cycle.")

        self.table.print_header(with_sep=False)
        self.stopped = False
        # Actual optimization loop
        for self.cur_cycle in range(self.last_cycle, self.max_cycles):
            start_time = time.time()
            self.log(highlight_text(f"Cycle {self.cur_cycle:03d}"))

            if self.is_cos and self.check_coord_diffs:
                image_coords = [image.cart_coords for image in self.geometry.images]
                align = len(image_coords[0]) > 3
                cds = get_coords_diffs(image_coords, align=align)
                # Differences of coordinate differences ;)
                cds_diffs = np.diff(cds)
                min_ind = cds_diffs.argmin()
                if cds_diffs[min_ind] < self.coord_diff_thresh:
                    similar_inds = min_ind, min_ind + 1
                    msg = (
                        f"Cartesian coordinates of images {similar_inds} are "
                        "too similar. Stopping optimization!"
                    )
                    # I should improve my logging :)
                    print(msg)
                    self.log(msg)
                    sim_fn = "too_similar_trj.xyz"
                    with open(sim_fn, "w") as handle:
                        handle.write(self.geometry.as_xyz())
                    print(f"Dumped latest coordinates to '{sim_fn}'.")
                    break

            # Check if something considerably changed in the optimization,
            # e.g. new images were added/interpolated. Then the optimizer
            # should be reset.
            reset_flag = False
            if self.cur_cycle > 0 and self.is_cos:
                reset_flag = self.geometry.prepare_opt_cycle(
                    self.coords[-1], self.energies[-1], self.forces[-1]
                )
            # Reset when number of coordinates changed
            elif self.cur_cycle > 0:
                reset_flag = reset_flag or (
                    self.geometry.coords.size != self.coords[-1].size
                )

            if reset_flag:
                self.reset()

            # Coordinates may be updated here.
            if self.reparam_when == "before" and hasattr(
                self.geometry, "reparametrize"
            ):
                # This call actually returns a bool, but right now we just drop it.
                self.geometry.reparametrize()

            self.coords.append(self.geometry.coords.copy())
            self.cart_coords.append(self.geometry.cart_coords.copy())

            # Determine and store number of currenctly actively optimized images
            try:
                image_inds = self.geometry.image_inds
                image_num = len(image_inds)
            except AttributeError:
                image_inds = [
                    0,
                ]
                image_num = 1
            self.image_inds.append(image_inds)
            self.image_nums.append(image_num)

            # Here the actual step is obtained from the actual optimizer class.
            step = self.optimize()
            step_norm = np.linalg.norm(step)
            self.log(f"norm(step)={step_norm:.6f} au (rad)")
            for source, target in (
                ("true_energy", "true_energies"),
                ("true_forces", "true_forces"),
            ):
                try:
                    if (value := getattr(self.geometry, source)) is not None:
                        getattr(self, target).append(value)
                except AttributeError:
                    pass

            if step is None:
                # Remove the previously added coords
                self.coords.pop(-1)
                self.cart_coords.pop(-1)
                continue

            if self.is_cos:
                self.tangents.append(self.geometry.get_tangents().flatten())

            self.steps.append(step)

            # Convergence check
            self.is_converged, conv_info = self.check_convergence()

            end_time = time.time()
            elapsed_seconds = end_time - start_time
            self.cycle_times.append(elapsed_seconds)

            if self.dump:
                self.write_cycle_to_file()
                with open(self.current_fn, "w") as handle:
                    handle.write(self.geometry.as_xyz())

            if (
                self.dump
                and self.dump_restart
                and (self.cur_cycle % self.dump_restart) == 0
            ):
                self.dump_restart_info()

            if (self.cur_cycle % self.print_every) == 0 or self.is_converged:
                self.print_opt_progress(conv_info)
            if self.is_converged:
                self.table.print("Converged!")
                break
            # Allow convergence, before checking for too small steps
            elif self.assert_min_step and (step_norm <= self.min_step_norm):
                raise ZeroStepLength

            # Update coordinates
            new_coords = self.geometry.coords.copy() + step
            try:
                self.geometry.coords = new_coords
                # Use the actual step. It may differ from the proposed step
                # when internal coordinates are used, as the internal-Cartesian
                # transformation is done iteratively.
                self.steps[-1] = self.geometry.coords - self.coords[-1]
            except RebuiltInternalsException as exception:
                print("Rebuilt internal coordinates!")
                rebuilt_fn = self.get_path_for_fn("rebuilt_primitives.xyz")
                with open(rebuilt_fn, "w") as handle:
                    handle.write(self.geometry.as_xyz())
                if self.is_cos:
                    for image in self.geometry.images:
                        image.reset_coords(exception.typed_prims)
                self.reset()

                if self.dump:
                    self.data_model = get_data_model(
                        self.geometry, self.is_cos, self.max_cycles
                    )
                    self.h5_group_name += "_rebuilt"
                    # h5_group = get_h5_group(
                    #     self.h5_fn,
                    #     self.h5_group_name,
                    #     self.data_model,
                    #     reset=True,
                    # )
                    # h5_group.file.close()

            # Coordinates may be updated here.
            if (self.reparam_when == "after") and hasattr(
                self.geometry, "reparametrize"
            ):
                reparametrized = self.geometry.reparametrize()
            else:
                reparametrized = False

            cur_coords = self.geometry.coords
            prev_coords = self.coords[-1]
            if (
                self.is_cos
                and self.reparam_check_rms
                and reparametrized
                and (cur_coords.size == prev_coords.size)
            ):
                self.log("Did reparametrization")

                rms = np.sqrt(np.mean((prev_coords - cur_coords) ** 2))
                self.log(f"rms of coordinates after reparametrization={rms:.6f}")
                self.is_converged = rms < self.reparam_thresh
                if self.is_converged:
                    self.table.print(
                        "Insignificant coordinate change after "
                        "reparametrization. Signalling convergence!"
                    )
                    break

            # Alternative: calcualte overlap of AFIR force and step. If this
            # overlap is negative the step is taken against the AFIR force.
            if self.monitor_frag_dists_counter > 0:
                interfrag_dist = interfragment_distance(
                    *self.fragments, self.geometry.coords3d
                )
                try:
                    prev_interfrag_dist = self.interfrag_dists[-1]
                    if interfrag_dist > prev_interfrag_dist:
                        print("Interfragment distances increased!")
                        self.stopped = True  # Can't use := in if clause
                        break
                except IndexError:
                    pass
                self.interfrag_dists.append(interfrag_dist)
                self.monitor_frag_dists_counter -= 1

            sys.stdout.flush()
            sign = check_for_end_sign()
            if sign == "stop":
                self.stopped = True
                break
            elif sign == "converged":
                self.converged = True
                self.table.print("Operator indicated convergence!")
                break

            self.log("")
        else:
            self.table.print("Number of cycles exceeded!")

        # Outside loop
        print()
        if self.dump:
            self.out_trj_handle.close()

        if (not self.is_cos) and (not self.stopped):
            print(self.final_summary())
            # Remove 'current_geometry.xyz' file
            try:
                os.remove(self.current_fn)
            except FileNotFoundError:
                self.log(f"Tried to delete '{self.current_fn}'. Couldn't find it.")
        with open(self.final_fn, "w") as handle:
            handle.write(self.geometry.as_xyz())
        print(f"Wrote final geometry to '{self.final_fn.name}'\n")
        self.postprocess_opt()
        sys.stdout.flush()

    def _get_opt_restart_info(self):
        """To be re-implemented in the derived classes."""
        return dict()

    def _set_opt_restart_info(self, opt_restart_info):
        """To be re-implemented in the derived classes."""
        return

    def get_restart_info(self):
        restart_info = {
            "geom_info": self.geometry.get_restart_info(),
            "last_cycle": self.cur_cycle,
            "max_cycles": self.max_cycles,
            "energies": self.energies,
            "coords": self.coords,
            "forces": [forces.tolist() for forces in self.forces],
            "steps": [step.tolist() for step in self.steps],
        }
        restart_info.update(self._get_opt_restart_info())
        return restart_info

    def set_restart_info(self, restart_info):
        # Set restart information general to all optimizers
        self.last_cycle = restart_info["last_cycle"] + 1

        must_resize = self.last_cycle >= self.max_cycles
        if must_resize:
            self.max_cycles += restart_info["max_cycles"]
            # Resize HDF5
            if self.dump:
                # h5_group = get_h5_group(self.h5_fn, self.h5_group_name)
                # resize_h5_group(h5_group, self.max_cycles)
                # h5_group.file.close()
                pass

        self.coords = [np.array(coords) for coords in restart_info["coords"]]
        self.energies = restart_info["energies"]
        self.forces = [np.array(forces) for forces in restart_info["forces"]]
        self.steps = [np.array(step) for step in restart_info["steps"]]

        # Set subclass specific information
        self._set_opt_restart_info(restart_info)

        # Propagate restart information downwards to the geometry
        self.geometry.set_restart_info(restart_info["geom_info"])

    def dump_restart_info(self):
        restart_info = self.get_restart_info()

        restart_fn = f"restart_{self.cur_cycle:03d}.yaml"
        restart_yaml = yaml.dump(restart_info)
        self.write_to_out_dir(restart_fn, restart_yaml)



try:
    import daltonproject as dp
except ModuleNotFoundError:
    print("daltonproject is not installed!")

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.constants import BOHR2ANG


class Dalton(Calculator):

    conf_key = "dalton"

    def __init__(self, basis, method="hf", **kwargs):
        super().__init__(**kwargs)

        self._basis = basis
        self._method = method

        self.basis = dp.Basis(basis=self._basis)
        self.method = dp.QCMethod(self._method)
        self.compute_settings = dp.ComputeSettings(mpi_num_procs=1)

    def prepare_input(self, atoms, coords):
        coords3d = coords.reshape(-1, 3) * BOHR2ANG
        dp_atoms = "; ".join(
            [f"{atom} {x} {y} {z}" for atom, (x, y, z) in zip(atoms, coords3d)]
        )
        molecule = dp.Molecule(atoms=dp_atoms, charge=self.charge)

        return molecule

    def compute(self, mol, prop):
        return dp.dalton.compute(
            mol, self.basis, self.method, prop, compute_settings=self.compute_settings
        )

    def get_energy(self, atoms, coords, **prepare_kwargs):
        mol = self.prepare_input(atoms, coords)
        prop = dp.Property(energy=True)
        res = self.compute(mol, prop)

        results = {
            "energy": res.energy,
        }
        return results


    def get_forces(self, atoms, coords, **prepare_kwargs):
        mol = self.prepare_input(atoms, coords)
        prop = dp.Property(energy=True, gradients=True)
        res = self.compute(mol, prop)

        results = {
            "energy": res.energy,
            "forces": -res.gradients.flatten(),
        }
        return results

    def __str__(self):
        return f"Dalton({self.name})"



import numpy as np

from pysisyphus.irc.IRC import IRC


# [1] http://pubs.acs.org/doi/pdf/10.1021/ja00295a002
# [2] https://math.stackexchange.com/questions/1882772/curve-fitting-with-derivatives


class IMKMod(IRC):

    def __init__(self, geometry, corr_first=True, corr_first_size=0.5,
                 corr_first_energy=True, corr_bisec_size=0.0025,
                 corr_bisec_energy=True, **kwargs):
        super().__init__(geometry, **kwargs)

        # Correction of pivot point
        self.corr_first = bool(corr_first)
        self.corr_first_size = float(corr_first_size)
        self.corr_first_energy = bool(corr_first_energy)
        # Correction along bisector
        self.corr_bisec_size = corr_bisec_size
        self.corr_bisec_energy = corr_bisec_energy

        self.corr_bisec_thresh = 2*self.corr_bisec_size

    def fit_parabola(self, x, y):
        y_str = np.array2string(np.array(y), precision=6)
        self.log(f"Fitting parabola with x={x} and y={y_str}")
        fit = np.polyfit(x, y, deg=2)
        parabola = np.poly1d(fit)
        minimum = parabola.deriv().r
        # We are only looking for a real minimum
        real_minimum = (minimum[minimum.imag==0].real)[0]
        self.log(f"Found minimum of fitted parabola at x={real_minimum:.4f}")

        return real_minimum

    def fit_grad_and_energies(self, energy0, grad0, energy1, direction):
        grad0_proj = grad0.dot(direction/np.linalg.norm(direction))
        # f (x) =  ax² + bx + c
        # f'(x) = 2ax  + b
        #
        # Assuming energy0 and grad0 are calculated at x=0
        # and
        # energy is calculated at x=1
        c = energy0
        b = grad0_proj
        a = energy1 - b - c
        coeffs = (a, b, c)
        poly = np.poly1d(coeffs)
        minimum = poly.deriv().r
        # We are only looking for a real minimum
        real_minimum = (minimum[minimum.imag==0].real)[0]
        self.log(f"Found minimum of fitted parabola at x={real_minimum:.4f}")

        return real_minimum

    def step(self):
        # Get gradient at Q0
        mw_coords_0 = self.mw_coords.copy()
        grad_0 = self.mw_gradient
        energy_0 = self.energy
        self.log(f"Current point:\n\tenergy_0={energy_0:.6f} au")


        ##############
        # PIVOT STEP #
        ##############

        grad_0_norm = np.linalg.norm(grad_0)
        dir_0 = -grad_0 / grad_0_norm
        # Step direction is against the gradient to the pivot point Q1.
        step = self.step_length * dir_0
        mw_coords_1 = self.mw_coords + step
        self.log(f"Took initial step of length {self.step_length:.6f} to pivot point")
        self.mw_coords = mw_coords_1
        energy_1 = self.energy

        #########################
        # PIVOT STEP CORRECTION #
        #########################

        self.log(f"Pivot point\n\tenergy_1={energy_1:.6f} au")
        energy_diff = energy_1 - energy_0
        if energy_diff > 0:
            self.log(f"Energy increased at pivot point ΔE={energy_diff:.6f}")
        correct_first_step = self.corr_first and (energy_diff > 0)
        # Fit using energy at q0, energy at q1, and energy at half step between
        # q0 and q1
        corr_min = None
        if correct_first_step and self.corr_first_energy:
            self.log("Computing correction for pivot point from energy_0, "
                     "energy_1, and additional energy calculation.")
            corr_step = self.corr_first_size*step
            mw_coords_1_corr = mw_coords_0 + corr_step
            self.mw_coords = mw_coords_1_corr
            energy_1_corr = self.energy
            x_corr = np.array((0, self.corr_first_size, 1))
            y_corr = (energy_0, energy_1_corr, energy_1)
            corr_min = self.fit_parabola(x_corr, y_corr)
        # Fit using energy & gradient at q0, and energy at q1
        elif correct_first_step:
            self.log("Computing correction for pivot point from energy_0, "
                     "grad_0 and energy_1, along dir_0")

            corr_min = self.fit_grad_and_energies(energy_0, grad_0, energy_1, dir_0)
        else:
            self.log("Skipping correction of pivot step")

        if corr_min:
            step_corr = corr_min * self.step_length * dir_0
            mw_coords_1 = mw_coords_0 + step_corr
            self.mw_coords = mw_coords_1

        # Get gradient at Q1
        self.log("Calculation of gradient at pivot point")
        grad_1 = self.mw_gradient
        energy_1 = self.energy
        grad_1_norm = np.linalg.norm(grad_1)
        dir_1 = -grad_1 / grad_1_norm

        ##############
        # BISECTOR D #
        ##############

        # Determine bisector d and normalized bisector vector D.
        # d = grad_0/grad_0_norm - grad_1/grad_1_norm
        d = -dir_0 + dir_1
        D = d / np.linalg.norm(d)
        self.log("Determined bisector D")

        ##########################
        # MINIMUM SEARCH ALONG D #
        ##########################

        # Take a step along D
        step_2 = self.corr_bisec_size * D
        mw_coords_2 = mw_coords_1 + step_2
        self.mw_coords = mw_coords_2
        energy_2 = self.energy
        self.log(f"1st step along D\n\tenergy_2={energy_2:.6f} au")

        if self.corr_bisec_energy:
            if energy_2 > energy_1:
                factor = 0.5
            else:
                factor = 2
            step_3 = self.corr_bisec_size * factor * D
            mw_coords_3 = mw_coords_1 + step_3
            self.mw_coords = mw_coords_3
            energy_3 = self.energy
            self.log(f"2nd step along D\n\tenergy_3={energy_3:.6f} au")
            x = np.array((0, 1, factor))
            y = (energy_1, energy_2, energy_3)
            corr_min = self.fit_parabola(x, y)
        else:
            self.log("Minimum search along D, by fitting energy_1, "
                     "grad_1 and energy_2.")
            corr_min = self.fit_grad_and_energies(energy_1, grad_1, energy_2, D)

        step_corr = self.corr_bisec_size * corr_min * D
        mw_coords_4 = mw_coords_1 + step_corr
        self.mw_coords = mw_coords_4



# pdb2reaction/backends/mace.py

"""
MACE (mace-torch) backend for pdb2reaction.

Requires: ``pip install mace-torch`` (separate env, conflicts with fairchem-core).
"""

from __future__ import annotations

import os
import tempfile
import urllib.request
import warnings
from typing import Optional, Sequence

import numpy as np

from .base import MLIPCalculator, BackendError

MACE_MP_ALIASES_FALLBACK = [
    "small", "medium", "large",
    "small-0b", "medium-0b",
    "small-0b2", "medium-0b2", "large-0b2",
    "medium-0b3", "medium-mpa-0",
    "small-omat-0", "medium-omat-0",
    "mace-matpes-pbe-0", "mace-matpes-r2scan-0",
    "mh-0", "mh-1",
]


class MACECalculator(MLIPCalculator):
    """MACE backend via mace.calculators."""

    def __init__(
        self,
        *,
        model: str = "MACE-OMOL-0",
        default_dtype: str = "float64",
        # Base class parameters
        charge: int = 0,
        spin: int = 1,
        device: str = "auto",
        freeze_atoms: Optional[Sequence[int]] = None,
        hessian_calc_mode: str = "FiniteDifference",
        return_partial_hessian: bool = False,
        hessian_double: bool = True,
        print_timing: bool = True,
        **kwargs,
    ):
        try:
            import torch
        except Exception as exc:
            raise BackendError(
                "MACE backend requires torch and mace-torch. "
                "Install with: pip install mace-torch (separate env required)"
            ) from exc

        # Warn about potential UMA/MACE conflict
        try:
            import fairchem.core  # noqa: F401
            warnings.warn(
                "fairchem-core is installed alongside mace-torch. "
                "These packages may conflict. If you encounter issues, "
                "use separate environments for UMA and MACE backends.",
                stacklevel=2,
            )
        except ImportError:
            pass

        super().__init__(
            charge=charge,
            spin=spin,
            device=device,
            freeze_atoms=freeze_atoms,
            hessian_calc_mode=hessian_calc_mode,
            return_partial_hessian=return_partial_hessian,
            hessian_double=hessian_double,
            print_timing=print_timing,
            **kwargs,
        )

        self._torch = torch
        if str(device).lower() == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device_str = str(device)

        self.model_spec = str(model)
        self.default_dtype = str(default_dtype)

        self._calc = self._build_calc(self.model_spec)

    @staticmethod
    def _download_to_tmp(url: str) -> str:
        target = os.path.join(
            tempfile.gettempdir(),
            os.path.basename(str(url)).split("?")[0] or "mace.model",
        )
        if not os.path.exists(target):
            urllib.request.urlretrieve(str(url), target)
        return target

    def _build_calc(self, model_spec: str):
        try:
            from mace.calculators import mace_anicc, mace_mp, mace_off, mace_omol
        except Exception as exc:
            raise BackendError(
                "Could not import mace.calculators. "
                "Install with: pip install mace-torch (separate env required)"
            ) from exc

        mp_aliases = []
        try:
            from mace.calculators.foundations_models import mace_mp_urls
            mp_aliases = list(mace_mp_urls.keys())
        except Exception:
            mp_aliases = list(MACE_MP_ALIASES_FALLBACK)

        spec = str(model_spec).strip()
        spec_l = spec.lower()
        mp_alias_lookup = {str(x).lower(): x for x in mp_aliases}

        def _mk_from_path(path_or_url):
            from mace.calculators.mace import MACECalculator as _MC
            model_path = str(path_or_url)
            if model_path.startswith("http://") or model_path.startswith("https://"):
                model_path = self._download_to_tmp(model_path)
            return _MC(model_paths=model_path, device=self.device_str, default_dtype=self.default_dtype)

        def _safe_anicc(kwargs):
            try:
                return mace_anicc(**kwargs)
            except Exception as exc:
                msg = str(exc).lower()
                if ("no nvidia driver" in msg) or ("cuda" in msg and "backend" in msg):
                    raise BackendError(
                        "MACE ANICC could not be loaded. Try a CUDA-enabled node or another MACE model."
                    ) from exc
                raise

        # Prefix forms
        if spec_l.startswith("mp:"):
            alias = spec.split(":", 1)[1].strip() or None
            if alias is not None:
                alias = mp_alias_lookup.get(str(alias).lower(), alias)
            return mace_mp(model=alias, device=self.device_str, default_dtype=self.default_dtype)

        if spec_l.startswith("off:"):
            alias = spec.split(":", 1)[1].strip() or None
            if alias is not None:
                alias = str(alias).lower()
            return mace_off(model=alias, device=self.device_str, default_dtype=self.default_dtype)

        if spec_l.startswith("omol:"):
            alias = spec.split(":", 1)[1].strip() or None
            if alias == "":
                alias = None
            if alias is not None:
                alias_l = str(alias).lower()
                if alias_l in ("mace-omol-0", "mace_omol_0", "maceomol0"):
                    alias = "extra_large"
                else:
                    alias = alias_l
            return mace_omol(model=alias, device=self.device_str, default_dtype=self.default_dtype)

        if spec_l.startswith("anicc"):
            path = None
            if ":" in spec:
                path = spec.split(":", 1)[1].strip() or None
            kwargs = {"device": self.device_str}
            if path:
                kwargs["model_path"] = path
            return _safe_anicc(kwargs)

        # Alias forms
        if spec_l in mp_alias_lookup:
            return mace_mp(model=mp_alias_lookup[spec_l], device=self.device_str, default_dtype=self.default_dtype)

        if spec_l in ("off-small", "off-medium", "off-large"):
            alias = spec_l.split("-", 1)[1]
            return mace_off(model=alias, device=self.device_str, default_dtype=self.default_dtype)

        if spec_l in ("omol-extra_large", "extra_large", "mace-omol-0", "mace_omol_0", "maceomol0"):
            return mace_omol(model="extra_large", device=self.device_str, default_dtype=self.default_dtype)

        if spec_l in ("anicc", "ani", "ani500k"):
            return _safe_anicc({"device": self.device_str})

        # Local file / URL
        if os.path.exists(spec) or spec.startswith("http://") or spec.startswith("https://"):
            return _mk_from_path(spec)

        raise BackendError(f"Unknown MACE model spec '{spec}'.")

    def _compute_energy_forces_ev(self, elem, coord_ang):
        from ase import Atoms

        atoms = Atoms(symbols=list(elem), positions=np.asarray(coord_ang, dtype=np.float64))
        atoms.info["charge"] = int(self.charge)
        atoms.info["spin"] = int(self.mult)
        atoms.calc = self._calc

        energy = float(atoms.get_potential_energy())
        forces = np.asarray(atoms.get_forces(), dtype=np.float64)
        return energy, forces


class MACEASECalculator:
    """Factory that returns a MACE ASE calculator for DMF."""

    def __new__(
        cls,
        *,
        model: str = "MACE-OMOL-0",
        device: str = "auto",
        default_dtype: str = "float64",
    ):
        calc = MACECalculator(
            model=model, device=device, default_dtype=default_dtype,
            charge=0, spin=1,
        )
        return calc._calc



# pdb2reaction/energy_diagram.py

"""
Minimal energy-diagram utility from numeric inputs.

Examples:
    pdb2reaction energy-diagram -i 0 12.5 4.3 -o energy.png
    pdb2reaction energy-diagram -i "[-205.1, -190.4, -198.7]" --label-x R TS P
"""

from __future__ import annotations

import ast
from pathlib import Path
from typing import List, Sequence

import click

from .utils import build_energy_diagram, ensure_dir

_IMG_EXTS = {".png", ".jpg", ".jpeg", ".svg", ".pdf"}


def _normalize_image_path(path: Path) -> Path:
    p = Path(path)
    if p.suffix:
        ext = p.suffix.lower()
        if ext not in _IMG_EXTS:
            raise click.BadParameter(
                f"Unsupported output extension '{p.suffix}'. "
                f"Supported: {', '.join(sorted(_IMG_EXTS))}"
            )
        return p
    return p.with_suffix(".png")


def _flatten_numeric_token(token: str) -> List[float]:
    txt = str(token).strip()
    if not txt:
        return []

    try:
        obj = ast.literal_eval(txt)
    except Exception:
        obj = txt

    def _to_float(x) -> float:
        try:
            return float(x)
        except Exception:
            raise click.BadParameter(
                f"Invalid numeric value in -i/--input: {x!r}"
            )

    if isinstance(obj, (int, float)):
        return [float(obj)]

    if isinstance(obj, (list, tuple)):
        out: List[float] = []
        for x in obj:
            out.append(_to_float(x))
        return out

    # Support comma-separated plain text like "0, 12.5, 4.3".
    if "," in txt:
        out: List[float] = []
        for part in txt.split(","):
            part = part.strip()
            if part:
                out.append(_to_float(part))
        return out

    return [_to_float(txt)]


def _parse_numeric_inputs(tokens: Sequence[str]) -> List[float]:
    vals: List[float] = []
    for tok in tokens:
        vals.extend(_flatten_numeric_token(tok))
    if len(vals) < 2:
        raise click.BadParameter(
            "Provide at least two numeric values with -i/--input."
        )
    return vals


def _flatten_label_token(token: str) -> List[str]:
    txt = str(token).strip()
    if not txt:
        return []

    try:
        obj = ast.literal_eval(txt)
    except Exception:
        obj = txt

    if isinstance(obj, (list, tuple)):
        return [str(x) for x in obj]
    if isinstance(obj, str) and "," in obj:
        return [x.strip() for x in obj.split(",") if x.strip()]
    return [str(obj)]


def _parse_label_x(tokens: Sequence[str]) -> List[str]:
    labels: List[str] = []
    for tok in tokens:
        labels.extend(_flatten_label_token(tok))
    return [str(x) for x in labels if str(x).strip()]


def _gather_variadic_values(
    ctx_args: List[str],
    flag_names: Sequence[str],
    all_flags: Sequence[str],
) -> List[str]:
    """Collect variadic positional values after *flag_names* from Click extra args.

    Stops at the next recognised flag or end-of-args.
    """
    names_set = set(flag_names)
    stop_set = set(all_flags)
    vals: List[str] = []
    i = 0
    while i < len(ctx_args):
        tok = ctx_args[i]
        if tok in names_set:
            j = i + 1
            while j < len(ctx_args) and ctx_args[j] not in stop_set:
                vals.append(ctx_args[j])
                j += 1
            i = j
        else:
            i += 1
    return vals


_ALL_FLAGS = (
    "-i", "--input",
    "-o", "--output",
    "--label-x", "--label-y",
    "-h", "--help", "--help-advanced",
)


@click.command(
    name="energy-diagram",
    help=(
        "Plot an energy diagram from numeric inputs only. "
        "Use -i for values (multiple numbers or list-like string)."
    ),
    context_settings={
        "help_option_names": ["-h", "--help"],
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    },
)
@click.option(
    "-i",
    "--input",
    "input_values",
    type=str,
    multiple=True,
    required=True,
    help=(
        "Numeric sequence. Accepts: "
        "-i 0 12.5 4.3  or  -i \"[0, 12.5, 4.3]\"  or repeated -i."
    ),
)
@click.option(
    "-o",
    "--output",
    "output_path",
    type=click.Path(path_type=Path, dir_okay=False),
    default=Path("energy_diagram.png"),
    show_default=True,
    help="Output image path.",
)
@click.option(
    "--label-x",
    "label_x",
    type=str,
    multiple=True,
    default=(),
    help=(
        "State labels on x-axis. Accepts: "
        "--label-x R TS P  or  --label-x \"['R','TS','P']\"."
    ),
)
@click.option(
    "--label-y",
    "label_y",
    type=str,
    default="ΔE (kcal/mol)",
    show_default=True,
    help="Y-axis label.",
)
@click.pass_context
def cli(
    ctx: click.Context,
    input_values: Sequence[str],
    output_path: Path,
    label_x: Sequence[str],
    label_y: str,
) -> None:
    # Click with ignore_unknown_options may not capture all variadic
    # positional values after -i / --label-x.  Re-parse from ctx.args
    # (the extra args Click could not match) to recover them.
    extra_inputs = _gather_variadic_values(
        ctx.args, ("-i", "--input"), _ALL_FLAGS,
    )
    input_tokens = list(input_values) + extra_inputs if extra_inputs else list(input_values)

    extra_labels = _gather_variadic_values(
        ctx.args, ("--label-x",), _ALL_FLAGS,
    )
    label_tokens = list(label_x) + extra_labels if extra_labels else list(label_x)

    energies = _parse_numeric_inputs(input_tokens)

    labels = _parse_label_x(label_tokens)
    if labels:
        if len(labels) != len(energies):
            raise click.BadParameter(
                f"--label-x count ({len(labels)}) must match value count ({len(energies)})."
            )
        labels_use = labels
    else:
        labels_use = [f"S{i + 1}" for i in range(len(energies))]

    out_img = _normalize_image_path(Path(output_path).resolve())
    ensure_dir(out_img.parent)

    fig = build_energy_diagram(
        energies=energies,
        labels=labels_use,
        ylabel=label_y,
        baseline=True,
        showgrid=False,
    )
    fig.write_image(str(out_img), scale=2)
    click.echo(f"[energy-diagram] Saved -> {out_img}")
